/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * data_scrollbar_controller - functions for managing a set of gtk
 * scrollbars for selecting a subrange from a given maximum range (for
 * example x axis of a plot).
 */

#include <glade/glade-xml.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <gtk/gtkrange.h>
#include <gtk/gtkversion.h>

#include "data_scrollbar_controller.xh"
#include "mem.xh"
#include "gtkutil.xh"
#include "util.xh"

#if 0
#define pprintf printf
#else
#define pprintf(...)
#endif

static void scroll_off_rel__value_changed(GtkRange *range, data_scrollbar_controller_t *this) {
   gtk_range_set_value(range, 0.0);
   this->scroll_off_rel_active = 0;
}

static gboolean scroll_off_rel__change_value(GtkRange *range, GtkScrollType scroll, gdouble value, data_scrollbar_controller_t *this) {
   if(!this->scroll_off_rel_active) {
      this->scroll_off_rel_active = 1;
      this->scroll_off_rel_pos = this->scroll_off->value;
   }
   if(scroll == GTK_SCROLL_JUMP) {
      x_gtk_adjustment_set_value_clamped(this->scroll_off, this->scroll_off_rel_pos + value * this->scroll_off->page_size);
   } else {
      double off;
      switch((int)scroll) {
	 case GTK_SCROLL_PAGE_FORWARD:  off =  1;    break;
	 case GTK_SCROLL_PAGE_BACKWARD: off = -1;    break;
	 case GTK_SCROLL_STEP_FORWARD:  off =  0.25; break;
	 case GTK_SCROLL_STEP_BACKWARD: off = -0.25; break;
	 default:
	    return 0;
      }
      x_gtk_adjustment_set_value_clamped(this->scroll_off, this->scroll_off->value + off * this->scroll_off->page_size);
   }

   return 0; // continue
}

static void scroll_off__value_changed(GtkAdjustment *adj, data_scrollbar_controller_t *this) {
   pprintf("%s\n", __FUNCTION__);

   pprintf("[%d] = %lf < %lf < %lf - %lf\n", __LINE__, this->scroll_off->lower, this->scroll_off->value, this->scroll_off->upper, this->zoompoint->value);

   if(this->label_off_min) {
      gtk_label_set_text(this->label_off_min,
			 format_time((my_time_t)this->scroll_off->value, 1));
   }
   if(this->label_off_max) {
      gtk_label_set_text(this->label_off_max,
			 format_time((my_time_t)(this->scroll_off->value + exp(this->scroll_scale->value)), 1));
   }

   this->num_changes++;
}

static void scroll_off__changed(GtkAdjustment *adj, data_scrollbar_controller_t *this) {
   double old_length = this->old_max - this->old_min;
   double new_length = this->scroll_off->upper - this->scroll_off->lower;

   if(this->label_off_max) {
      gtk_label_set_text(this->label_off_max,
			 format_time((my_time_t)(this->scroll_off->value + exp(this->scroll_scale->value)), 1));
   }

   pprintf("%s %lf\n", __FUNCTION__, new_length);

   pprintf("[%d] = %lf < %lf < %lf - %lf\n", __LINE__, this->scroll_off->lower, this->scroll_off->value, this->scroll_off->upper, this->zoompoint->value);

   double old_zoompoint = this->zoompoint->value;
   double new_zoompoint = (this->old_min + old_zoompoint * old_length - this->scroll_off->lower) / new_length;

   pprintf("[%d] = %lf %lf %lf %lf %lf\n", __LINE__, this->old_min, old_zoompoint, old_length, this->scroll_off->lower, new_length);

   x_gtk_adjustment_set_value_clamped(this->zoompoint, new_zoompoint);

   pprintf("[%d] = %lf < %lf < %lf - %lf\n", __LINE__, this->scroll_off->lower, this->scroll_off->value, this->scroll_off->upper, this->zoompoint->value);

   x_gtk_adjustment_set_range(this->scroll_scale, log(this->min_interval), log(new_length), 1);

   this->old_max = this->scroll_off->upper;
   this->old_min = this->scroll_off->lower;

   pprintf("[%d] = %lf < %lf < %lf - %lf\n", __LINE__, this->scroll_off->lower, this->scroll_off->value, this->scroll_off->upper, this->zoompoint->value);

   this->num_changes++;
}

static void scroll_scale__value_changed(GtkAdjustment *adj, data_scrollbar_controller_t *this) {
   double new_scroll_scale = exp(this->scroll_scale->value);

   pprintf("%s %lf / %lf\n", __FUNCTION__, this->scroll_scale->value, new_scroll_scale);

   pprintf("[%d] = %lf < %lf < %lf - %lf\n", __LINE__, this->scroll_off->lower, this->scroll_off->value, this->scroll_off->upper, this->zoompoint->value);

   this->scroll_off->page_size = new_scroll_scale;

   // adjust off
   double old_scroll_scale = this->old_scroll_scale;
   double old_scroll_off = this->scroll_off->value;
   double old_zoompoint = this->zoompoint->value;

   pprintf("[%d] = %lf %lf %lf %lf\n", __LINE__, old_scroll_off, new_scroll_scale, old_scroll_scale, old_zoompoint);

   double new_scroll_off = old_scroll_off - (new_scroll_scale - old_scroll_scale) * old_zoompoint;
   x_gtk_adjustment_set_value_clamped(this->scroll_off, new_scroll_off);

   this->old_scroll_scale = new_scroll_scale;

   // if zoom out caused offset to go outside range (which was clamped by x_gtk_adjustment_set_value_clamped), adjust zoompoint pos
   if(new_scroll_off < this->scroll_off->lower) {
      double new_zoompoint = old_zoompoint + (new_scroll_off - this->scroll_off->lower) / new_scroll_scale;
      x_gtk_adjustment_set_value_clamped(this->zoompoint, new_zoompoint);
   }
   if(new_scroll_off + new_scroll_scale > this->scroll_off->upper) {
      double new_zoompoint = old_zoompoint + (new_scroll_off + new_scroll_scale - this->scroll_off->upper) / new_scroll_scale;
      x_gtk_adjustment_set_value_clamped(this->zoompoint, new_zoompoint);
   }

   // tell the guys we are finished changing - we do this after setting value to do it in the same way GtkRange does.
   gtk_adjustment_changed(this->scroll_off);

   // ##TODO## assumes scrollbar unit is time
   if(this->entry_scale && !GTK_WIDGET_HAS_FOCUS(this->entry_scale)) {
      if(GTK_IS_ENTRY(this->entry_scale)) {
	 gtk_entry_set_text(GTK_ENTRY(this->entry_scale), format_difftime((my_time_t)new_scroll_scale, 1));
      } else {
	 gtk_label_set_text(GTK_LABEL(this->entry_scale), format_difftime((my_time_t)new_scroll_scale, 1));
      }
   }

   pprintf("[%d] = %lf < %lf < %lf - %lf\n", __LINE__, this->scroll_off->lower, this->scroll_off->value, this->scroll_off->upper, this->zoompoint->value);

   this->num_changes++;
}

static void entry_scale__activate(GtkEntry *entry, data_scrollbar_controller_t *this) {

/* ##TODO##
   my text = this->entry_scale->get_text();
   my newval = 0;
   if(text =~ m!^\s*(\d+)\s*d(?:ays?)?(.*)!) {
      newval += $1 * 86400;
      text = $2;
   }
   if(text =~ m!^\s*(\d+)\s*h(?:ours?)?(.*)!) {
      newval += $1 * 3600;
      text = $2;
   }
   if(text =~ m!^\s*(\d+)\s*m(?:in(?:ute)?s?)?(.*)!) {
      newval += $1 * 60;
      text = $2;
   }
   if(text =~ m!^\s*(\d+)\s*s(?:ec(?:ond)?s?)?(.*)!) {
      newval += $1;
      text = $2;
   }
   if(text =~ m!^\s*(\d+)\s*:\s*(\d+)\s*:\s*(\d+)\s*!) {
      newval += $1 * 3600 + $2 * 60 + $3;
   }
   my min = this->min_interval};
    my max = this->scroll_off->upper} - this->scroll_off->lower};
   newval = min if(newval < min);
   newval = max if(newval > max);
   this->set_page_length(newval);

   // below is copypasted from above

   my s = newval % 60; newval = int(newval / 60);
   my m = newval % 60; newval = int(newval / 60);
   my h = newval % 24; newval = int(newval / 24);

   this->entry_scale->set_text(sprintf("%d days %02d:%02d:%02d", newval, h, m, s));
*/
}

static GtkAdjustment *get_or_create_adjustment(GtkRange *range, double value, double lower, double upper, double step_increment, double page_increment, double page_size) {
   if(range) {
      GtkAdjustment *adj = gtk_range_get_adjustment(range);
      g_object_ref(G_OBJECT(adj));
      return adj;
   } else {
      GtkAdjustment *adj = GTK_ADJUSTMENT(gtk_adjustment_new(value, lower, upper, step_increment, page_increment, page_size));
      g_object_ref(adj);
      gtk_object_sink(GTK_OBJECT(adj));
      return adj;
   }
}

static void reset_old(data_scrollbar_controller_t *this) {
   this->old_scroll_scale = this->scroll_scale->value;
   this->old_min = this->scroll_off->lower;
   this->old_max = this->scroll_off->upper;
}

static void connect_signals(data_scrollbar_controller_t *this) {
   // off changed
   g_signal_connect(this->scroll_off, "value-changed", G_CALLBACK(scroll_off__value_changed), this);
   g_signal_connect(this->scroll_off, "changed", G_CALLBACK(scroll_off__changed), this);

   // scale changed
   g_signal_connect(this->scroll_scale, "value-changed", G_CALLBACK(scroll_scale__value_changed), this);

   if(this->entry_scale) {
      // if user edits the entry_scale field and hits enter, try to parse it!
      g_signal_connect(this->entry_scale, "activate", G_CALLBACK(entry_scale__activate), this);
   }
}

data_scrollbar_controller_t *data_scrollbar_controller_new(GladeXML *gladexml, const char *widget_prefix) {
   data_scrollbar_controller_t *this = zmalloc_c(sizeof(data_scrollbar_controller_t));

   GtkRange *scroll_scale_range;
   GtkRange *scroll_off_range;
   GtkRange *scroll_off_rel_range;
   GtkRange *zoompoint_range;

   GtkWidget *entry_scale;
   GtkLabel *label_off_min;
   GtkLabel *label_off_max;

   if(gladexml && widget_prefix) {
      size_t wpl = strlen(widget_prefix);
      char tmp[wpl + 18];
      memcpy(tmp, widget_prefix, wpl);

      scroll_scale_range   = GTK_RANGE(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "scrollbar_scale"), tmp)));
      scroll_off_range     = GTK_RANGE(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "scrollbar_off"), tmp)));
      scroll_off_rel_range = GTK_RANGE(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "scrollbar_off_rel"), tmp)));
      zoompoint_range      = GTK_RANGE(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "scale"), tmp)));

      entry_scale    = glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "entry_scale"), tmp));
      label_off_min  = GTK_LABEL(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "label_off_min"), tmp)));
      label_off_max  = GTK_LABEL(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "label_off_max"), tmp)));
   } else {
      scroll_scale_range   = NULL;
      scroll_off_range     = NULL;
      scroll_off_rel_range = NULL;
      zoompoint_range      = NULL;

      entry_scale    = NULL;
      label_off_min  = NULL;
      label_off_max  = NULL;
   }

   this->scroll_scale = get_or_create_adjustment(scroll_scale_range, 1.,0.,1.,0.,0.,0.);
   this->scroll_off = get_or_create_adjustment(scroll_off_range, 0.,0.,1.,0.,0.,0.);
   this->zoompoint = get_or_create_adjustment(zoompoint_range, 0.5,0.,1.,0.,0.,0.1);
   this->entry_scale = entry_scale;
   this->label_off_min = label_off_min;
   this->label_off_max = label_off_max;

   this->scroll_scale->step_increment = log(2.0) / 10; // double/halves the scale in 10 presses if "step" change
   this->scroll_scale->page_increment = log(2.0); // doubles/halves the scale if "page" change
   this->scroll_off->step_increment = 1000.0; // ##TODO## configure
   this->scroll_off->page_increment = 600000.0;

   this->min_interval = 10;
   this->scroll_scale->page_size = 0;

   if(scroll_off_rel_range) {

#if GTK_CHECK_VERSION(2,6,0)

      // by using discontinuous mode, the value_changed is just
      // called once after a mouse scrollbar drag while change_value
      // is called during the drag
      gtk_range_set_update_policy(scroll_off_rel_range, GTK_UPDATE_DISCONTINUOUS);

      // this is called when user lets go of the scrollbar
      g_signal_connect(scroll_off_rel_range, "value-changed", G_CALLBACK(scroll_off_rel__value_changed), this);

      // this is called when the user scrolls
      g_signal_connect(scroll_off_rel_range, "change-value", G_CALLBACK(scroll_off_rel__change_value), this);

#else

      // change-value signal not supported, so this widget is useless.. hide it
      gtk_widget_hide(GTK_WIDGET(scroll_off_rel_range));

#endif

   }

   reset_old(this);
   connect_signals(this);

   this->scroll_scale->lower = 0.01;
   this->scroll_scale->value = 0.01;
   this->scroll_scale->upper = 0.01;
   this->scroll_off->lower = 0;
   this->scroll_off->value = 0;
   this->scroll_off->upper = 10;

   gtk_adjustment_changed(this->scroll_scale);
   gtk_adjustment_value_changed(this->scroll_scale);
   gtk_adjustment_changed(this->scroll_off);
   gtk_adjustment_value_changed(this->scroll_off);

   return this;
}

data_scrollbar_controller_t *data_scrollbar_controller_clone(data_scrollbar_controller_t *orig) {
   data_scrollbar_controller_t *this = zmalloc_c(sizeof(data_scrollbar_controller_t));

   this->scroll_scale = x_gtk_adjustment_clone(orig->scroll_scale);
   this->scroll_off = x_gtk_adjustment_clone(orig->scroll_off);
   this->zoompoint = x_gtk_adjustment_clone(orig->zoompoint);

   this->min_interval = orig->min_interval;

   reset_old(this);
   connect_signals(this);

   return this;
}

void data_scrollbar_controller_copy(data_scrollbar_controller_t *n, data_scrollbar_controller_t *o) {
   x_gtk_adjustment_copy(n->scroll_scale, o->scroll_scale);
   x_gtk_adjustment_copy(n->scroll_off, o->scroll_off);
   x_gtk_adjustment_copy(n->zoompoint, o->zoompoint);

   n->min_interval = o->min_interval;

   reset_old(n);

   gtk_adjustment_changed(n->zoompoint);
   gtk_adjustment_value_changed(n->zoompoint);
   gtk_adjustment_changed(n->scroll_scale);
   gtk_adjustment_value_changed(n->scroll_scale);
   gtk_adjustment_changed(n->scroll_off);
   gtk_adjustment_value_changed(n->scroll_off);
}

void data_scrollbar_controller_free(data_scrollbar_controller_t *this) {
   g_object_unref(this->scroll_off);
   g_object_unref(this->scroll_scale);
   g_object_unref(this->zoompoint);

   bzero(this, sizeof(data_scrollbar_controller_t));
   free(this);
}

void data_scrollbar_controller_set_page_off(data_scrollbar_controller_t *this, double off) {
   x_gtk_adjustment_set_value_clamped(this->scroll_off, off);
}

double data_scrollbar_controller_get_page_off(data_scrollbar_controller_t *this) {
   return this->scroll_off->value;
}

void data_scrollbar_controller_set_page_length(data_scrollbar_controller_t *this, double length) {
   if(length > this->scroll_off->upper - this->scroll_off->lower) {
      length = this->scroll_off->upper - this->scroll_off->lower;
   } else if(length < this->min_interval) {
      length = this->min_interval;
   }
   pprintf("scroll = %lf log = %lf\n", length, log(length));
   x_gtk_adjustment_set_value_clamped(this->scroll_scale, log(length));
}

void data_scrollbar_controller_set_page_length_log(data_scrollbar_controller_t *this, double loglength) {
   if(loglength > log(this->scroll_off->upper - this->scroll_off->lower)) {
      loglength = log(this->scroll_off->upper - this->scroll_off->lower);
   } else if(loglength < log(this->min_interval)) {
      loglength = log(this->min_interval);
   }
   pprintf("scroll log = %lf\n", loglength);
   x_gtk_adjustment_set_value_clamped(this->scroll_scale, loglength);
}

double data_scrollbar_controller_get_page_length(data_scrollbar_controller_t *this) {
   return exp(this->scroll_scale->value);
}

double data_scrollbar_controller_get_page_length_log(data_scrollbar_controller_t *this) {
   return this->scroll_scale->value;
}

void data_scrollbar_controller_set_range(data_scrollbar_controller_t *this, double lo, double hi) {
   // this triggers scroll_off__changed which updates this->scroll_scale range
   x_gtk_adjustment_set_range(this->scroll_off, lo, hi, 0);
}

void data_scrollbar_controller_get_range(data_scrollbar_controller_t *this, double *lo, double *hi) {
   *lo = this->scroll_off->lower;
   *hi = this->scroll_off->upper;
}

void data_scrollbar_controller_set_min_interval(data_scrollbar_controller_t *this, double min_interval) {
   this->min_interval = min_interval;
   pprintf("scroll mininterval %lf\n", min_interval);
   x_gtk_adjustment_set_range(this->scroll_scale, log(min_interval), log(this->scroll_off->upper - this->scroll_off->lower), 1);
}

double data_scrollbar_controller_get_min_interval(data_scrollbar_controller_t *this) {
   return this->min_interval;
}

void data_scrollbar_controller_set_zoompoint(data_scrollbar_controller_t *this, double zoompoint) {
   if(zoompoint < 0) {
      printf("warning: zoompoint < 0 (%lf)\n", zoompoint);
      zoompoint = 0;
   } else if(zoompoint > 1) {
      printf("warning: zoompoint > 1 (%lf)\n", zoompoint);
      zoompoint = 1;
   }
   x_gtk_adjustment_set_value_clamped(this->zoompoint, zoompoint);
}

double data_scrollbar_controller_get_zoompoint(data_scrollbar_controller_t *this) {
   return this->zoompoint->value;
}
