/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * plot2 - most functions related to representing & managing the
 * actual graphs along with gtk event handlers - except for the
 * plotting itself, see plot_draw
 */

#include <ctype.h>
#include <stdio.h>
#include <gtk/gtkdrawingarea.h>
#include <gtk/gtkfilechooserdialog.h>
#include <gtk/gtkfontsel.h>
#include <gtk/gtktable.h>
#include <gtk/gtkcombobox.h>
#include <gtk/gtkcheckmenuitem.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkcheckbutton.h>
#include <gtk/gtkversion.h>
#include <gtk/gtkscale.h>
#include <gtk/gtkrange.h>
#include <glib/giochannel.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "gnu_extensions.h"
#include "stringutil.xh"
#include "grid.xh"
#include "mem.xh"
#include "util.xh"
#include "gtkutil.xh"
#include "plot2.xh"
#include "export.xh"
#include "plot_draw.xh"
#include "array.xh"
#include "main.xh"
#include "whls.xh"
#include "compat.xh"
#include "paintapi-gtk.xh"

static void plot_need_redraw(my_plot_t *plot);
static void plot_view_need_redraw(my_plot_view_t *view);
static void pan_right(my_plot_t *plot);
static void plot_view_connect_signals(my_plot_view_t *view, GladeXML *gladexml);

// ##TODO## make local
// ##TODO## the colors suck
paintapi_rgb_t default_line_colors[] = {
   { 0xffff, 0x4444, 0x4444 },
   { 0xffff, 0xc0c0, 0x4040 },
   { 0x4040, 0xffff, 0x4040 },
   { 0x4040, 0xffff, 0xffff },
   { 0x4040, 0x8080, 0xffff },
   { 0xffff, 0x0000, 0x0000 },
   { 0x0000, 0xffff, 0x0000 },
   { 0x0000, 0x0000, 0xffff },
   { 0xffff, 0xffff, 0xffff },
};

static void about_clicked(GtkMenuItem *about_menuitem, my_plot_t *plot) {
   glade_xml_new(plot->gladefile, "aboutdialog1", NULL);
}

const char *labelNames[] = {
   "File date:",
   "Start time:",
   "End time:",
   "Covered time:",
   "Time format:",
   "Data columns:",
};

typedef struct {
      GtkWidget *da;
      GtkLabel *labels[SZ(labelNames)];
      paintapi_t *paintapi;
      my_plot_draw_t draw;
} adf_data_t;

static int get_black_color(GdkColor *dst);

static void adf_update(GtkFileChooser *chooser, adf_data_t *data) {
   char *filename = gtk_file_chooser_get_preview_filename(chooser);
   my_dataset_t *dataset = scan_dataset(filename, 10);
   gtk_file_chooser_set_preview_widget_active(chooser, 1);

   plot_draw_remove_all_data_files(&data->draw);

   if(!dataset) {
      struct stat st;
      if(!stat(filename, &st)) {
	 gtk_label_set_text(data->labels[0], format_time(st.st_mtime * TIME_MULTIPLIER, 0));
      } else {
	 gtk_label_set_text(data->labels[0], "xxxx-xx-xx xx:xx:xx");
      }
      gtk_label_set_text(data->labels[1], "");
      gtk_label_set_text(data->labels[2], "");
      gtk_label_set_text(data->labels[3], "");
      gtk_label_set_text(data->labels[4], "");
      gtk_label_set_text(data->labels[5], "");
      gtk_widget_queue_draw(data->da);

      g_free(filename);
      return;
   }

   char buf[40];
   int idx = 0;

   gtk_label_set_text(data->labels[idx++], format_time(dataset->last_modified, 0));
   gtk_label_set_text(data->labels[idx++], format_time(dataset->ts, 0));
   gtk_label_set_text(data->labels[idx++], format_time(dataset->te, 0));
   gtk_label_set_text(data->labels[idx++], format_difftime(dataset->te - dataset->ts, 0));
   gtk_label_set_text(data->labels[idx++], dataset->timetype == TIMETYPE_EPOCH ? "epoch" : "relative");
   snprintf(buf, sizeof(buf), "%d", dataset->item_count);
   gtk_label_set_text(data->labels[idx++], buf);

   unref_dataset(dataset);

   // read preview

   dataset = read_dataset_preview(filename, 400);

   g_free(filename);

   if(!dataset) {
      return;
   }

   my_plot_line_info_set_arr_ensure(data->draw.line_set, 1);

   data->draw.line_set[0].dataset = dataset;

   data->draw.xmin = data->draw.xstart = dataset->ts;
   data->draw.xmax = data->draw.xend = dataset->te;

   double vmin = INFINITY, vmax = -INFINITY;

   for(int dsi=0; dsi<dataset->item_count; ++dsi) {
      int li = data->draw.line_set[0].line_count;
      my_plot_line_info_arr_ensure(data->draw.line_set[0].line, li+1);
      int linecoloridx = dsi;
      data->draw.line_set[0].line[li].color = default_line_colors[linecoloridx % SZ(default_line_colors)];
      data->draw.line_set[0].line[li].dataset_idx = dsi;
      data->draw.line_set[0].line[li].enabled = 1;
      data->draw.line_set[0].line[li].mode = LINE_MODE_NORMAL;
      if(dsi == 0 || vmin > dataset->item[dsi].min) vmin = dataset->item[dsi].min;
      if(dsi == 0 || vmax < dataset->item[dsi].max) vmax = dataset->item[dsi].max;
   }

   data->draw.ymin = vmin - (vmax-vmin)/10;
   data->draw.ymax = vmax + (vmax-vmin)/10;

   gtk_widget_queue_draw(data->da);
}

static gboolean adtd_map_event(GtkWidget *widget, GdkEvent *do_not_use, adf_data_t *data) {
   data->draw.paintapi = paintapi_gtk_new(data->da->window, &data->draw);
      
   // set the background part 1 - create new style for widget
   GtkStyle *style = gtk_widget_get_style(data->da);
   style = gtk_style_copy(style);
   gtk_widget_set_style(data->da, style);
   
   // set the background part 2
   GdkColor bg;
   get_black_color(&bg);
   gdk_window_set_background(data->da->window, &bg);

   // fonts
   data->draw.font_legend = zmalloc_c(sizeof(paintapi_font_t));
   data->draw.font_xaxis = data->draw.font_legend;
   data->draw.font_yaxis = data->draw.font_legend;
   data->draw.font_legend->ctx = gtk_widget_get_pango_context(data->da);
   data->draw.font_legend->font = pango_font_description_from_string("Arial 9");
}

static gboolean adtd_configure_event(GtkWidget *widget, GdkEventConfigure *do_not_use, adf_data_t *data) {
   int w,h;
   gdk_drawable_get_size(data->da->window, &w, &h);
   data->draw.plot_xe = w - 1;
   data->draw.plot_ye = h - 1;
   return 0;
}

static gboolean adtd_expose_event(GtkWidget *widget, GdkEventExpose *event, adf_data_t *data) {
   if(data->draw.paintapi && data->draw.line_set_count)
      plot_draw(&data->draw);
   return 1;
}

static void add_data_file_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {

   GladeXML *gladexml = glade_xml_new(plot->gladefile, "filechooserdialog_add_file", NULL);
   GtkWidget *dialog = glade_xml_get_widget(gladexml, "filechooserdialog_add_file");

   GtkFileChooser *chooser = GTK_FILE_CHOOSER(dialog);

   gtk_file_chooser_set_select_multiple(chooser, 1);

   GtkWidget *table = gtk_table_new(5, 2, 0);
   gtk_table_set_col_spacings(GTK_TABLE(table), 2);

#define setcolor(x,r_,g_,b_) x.r=r_;x.g=g_;x.b=b_;

   adf_data_t adtd;
   bzero(&adtd, sizeof(adf_data_t));

   // plot_draw init
   plot_draw_init(&adtd.draw);

   setcolor(adtd.draw.minor_color, 0x6060, 0x6060, 0x6060);
   setcolor(adtd.draw.major_color, 0xc0c0, 0xc0c0, 0xc0c0);
   setcolor(adtd.draw.zerogrid_color, 0xffff, 0x6060, 0x4040);
   setcolor(adtd.draw.sep_color, 0x8080, 0x8080, 0x8080);
   setcolor(adtd.draw.label_color, 0xffff, 0xffff, 0xffff);
   setcolor(adtd.draw.cursor_marker_color, 0x8080, 0x8080, 0x8080);
   setcolor(adtd.draw.legend_bg_color, 0x2020, 0x2020, 0x2020);
   setcolor(adtd.draw.legend_border_color, 0xa0a0, 0xa0a0, 0xa0a0);
   setcolor(adtd.draw.legend_text_color, 0xe0e0, 0xe0e0, 0xe0e0);

   adtd.draw.show_legend = LEGEND_HIDE;

   // /plot_draw init

   adtd.da = gtk_drawing_area_new();
   gtk_widget_set_size_request(adtd.da, 80, 80);
   gtk_table_attach(GTK_TABLE(table), adtd.da, 0, 2, 0, 1, GTK_EXPAND|GTK_FILL, GTK_EXPAND|GTK_FILL, 0, 0);
   g_signal_connect(adtd.da, "map-event", G_CALLBACK(adtd_map_event), &adtd);
   g_signal_connect(adtd.da, "configure-event", G_CALLBACK(adtd_configure_event), &adtd);
   g_signal_connect(adtd.da, "expose-event", G_CALLBACK(adtd_expose_event), &adtd);

   for(unsigned i=0; i<SZ(labelNames); ++i) {
      GtkWidget *labelNameW = gtk_label_new(labelNames[i]);
      GtkWidget *labelW = gtk_label_new("?");

      gtk_misc_set_alignment(GTK_MISC(labelNameW), 1.0, 0.5);
      gtk_misc_set_alignment(GTK_MISC(labelW), 0.0, 0.5);

      adtd.labels[i] = GTK_LABEL(labelW);

      gtk_table_attach(GTK_TABLE(table), labelNameW, 0, 1, i+1, i+2, GTK_EXPAND|GTK_FILL, 0, 0, 0);
      gtk_table_attach(GTK_TABLE(table), labelW,     1, 2, i+1, i+2, GTK_EXPAND|GTK_FILL, 0, 0, 0);
   }
   gtk_widget_show_all(table);

   gtk_file_chooser_set_preview_widget(chooser, table);
   g_signal_connect(chooser, "update-preview", G_CALLBACK(adf_update), &adtd);

   GtkFileFilter *filter;

   filter = gtk_file_filter_new();
   gtk_file_filter_set_name(filter, "Text files");
   gtk_file_filter_add_mime_type(filter, "text/plain");
   gtk_file_chooser_add_filter(chooser, filter);

   filter = gtk_file_filter_new();
   gtk_file_filter_set_name(filter, "All files");
   gtk_file_filter_add_pattern(filter, "*");
   gtk_file_chooser_add_filter(chooser, filter);

   if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
      gtk_widget_hide(dialog);

      GSList* files;
      for(files = gtk_file_chooser_get_filenames(chooser); files; files = files->next) {
	 add_datafile(plot, files->data);
      }
      g_slist_free(files);
   }

   plot_draw_deinit(&adtd.draw);

   gtk_widget_destroy(dialog);
}

void remove_data_file(my_plot_t *plot, int idx) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot_draw_remove_data_file(&plot->view[vi]->draw, idx);
   }

   plot_need_redraw(plot);
   update_minmax(plot);
}

void remove_all_data_files(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot_draw_remove_all_data_files(&plot->view[vi]->draw);
   }

   plot_need_redraw(plot);
   update_minmax(plot);
}

static void reset_gc(my_plot_t *plot) {
   GdkColor bg;
   get_color(plot, &bg, "#000000");

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot_draw_reset_gc(&plot->view[vi]->draw);
      gdk_window_set_background(plot->view[vi]->drawable, &bg);
   }
}

static void remove_data_file_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   GladeXML *gladexml = glade_xml_new(plot->gladefile, "dialog_remove_datafile", NULL);
   GtkWidget *dialog = glade_xml_get_widget(gladexml, "dialog_remove_datafile");

   GtkWidget *combo = glade_xml_get_widget(gladexml, "combobox_datafiles");

   gtk_combo_box_remove_text(GTK_COMBO_BOX(combo), 0);

   for(int lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
      gtk_combo_box_append_text(GTK_COMBO_BOX(combo), plot->view[0]->draw.line_set[lsi].dataset->path);
   }

   if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
      gtk_widget_hide(dialog);

      int idx = gtk_combo_box_get_active(GTK_COMBO_BOX(combo));
      if(idx >= 0) {
#if GTK_CHECK_VERSION(2,6,0)
	 printf("removing %d. %s\n", idx, gtk_combo_box_get_active_text(GTK_COMBO_BOX(combo)));
#else
	 printf("removing %d.\n", idx);
#endif
	 remove_data_file(plot, idx);
      }
   }

   gtk_widget_destroy(dialog);
}

static void remove_all_data_files_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   remove_all_data_files(plot);
}

static void refresh_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   for(int lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
      update_dataset(plot->view[0]->draw.line_set[lsi].dataset);
   }

   plot_need_redraw(plot);
   update_minmax(plot);
}

#ifdef USE_FAM

static void FAMError(const char *msg) {
   fprintf(stderr, "%s: %s\n", msg, FamErrlist[FAMErrno]);
}

static gboolean fam_callback(GIOChannel *source, GIOCondition condition, my_plot_t *plot) {
   boolean_t updated = 0;
   while(FAMPending(&plot->fc)) {
      FAMEvent fe;
      //printf("FAM event\n");
      if(FAMNextEvent(&plot->fc, &fe) == -1) {
	 FAMError("fammonitorfile");
	 exit(1);
      }

      if(fe.code != FAMChanged) continue;

      //printf("FAM event changed\n");

      // TODO check only the file that the FAMChanged event indicates

      for(int lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
	 if(!plot->view[0]->draw.line_set[lsi].dataset->fam_monitored) continue;
	 if(plot->view[0]->draw.line_set[lsi].dataset->fr.reqnum != fe.fr.reqnum) continue;
	 int oldcount = plot->view[0]->draw.line_set[lsi].dataset->time_count;
	 if(!update_dataset(plot->view[0]->draw.line_set[lsi].dataset)) {
	    printf("Problems updating %s\n", plot->view[0]->draw.line_set[lsi].dataset->path);
	    // TODO maybe ask user whether to start from beginning
	    printf("remove watch for %s\n", plot->view[0]->draw.line_set[lsi].dataset->path);
	    if(FAMCancelMonitor(&plot->fc, &plot->view[0]->draw.line_set[lsi].dataset->fr)) {
	       fprintf(stderr, "Error cancelling fam monitor for %s: %s\n", plot->view[0]->draw.line_set[lsi].dataset->path, FamErrlist[FAMErrno]);
	    }
	    plot->view[0]->draw.line_set[lsi].dataset->fam_monitored = 0;
	 }
	 if(oldcount < plot->view[0]->draw.line_set[lsi].dataset->time_count) {
	    updated = 1;
	 }
      }
   }

   if(updated) {
      //printf("FAM updated\n");
      plot_need_redraw(plot);
      update_minmax(plot);
      pan_right(plot);
      // avoid zoompoint gliding implied by the zoompoint-maintaining algorithm in data_scrollbar_controller
      data_scrollbar_controller_set_zoompoint(plot->dsc_x, 1.0);
   }

   return 1;
}

static void enable_fam(my_plot_t *plot) {
   if(!plot->fam_initialized) {
      if(FAMOpen(&plot->fc)) {
	 FAMError("fammonitorfile");
	 return;
      }

      int fd = FAMCONNECTION_GETFD(&plot->fc);

      GIOChannel *fam_giochannel = g_io_channel_unix_new(fd);
      if(!fam_giochannel) {
	 FAMClose(&plot->fc);
	 return;
      }
      plot->fam_evtsrcid = g_io_add_watch(fam_giochannel, G_IO_IN, (GIOFunc)fam_callback, plot);
      g_io_channel_unref(fam_giochannel);
      plot->fam_initialized = 1;
   }

   gtk_widget_set_sensitive(plot->xscroll_off, 0);
   gtk_widget_set_sensitive(plot->xscroll_off_rel, 0);
   gtk_widget_set_sensitive(plot->xscroll_zoompoint, 0);
   gtk_widget_set_sensitive(plot->menuitem_file_refresh, 0);

   pan_right(plot);

   data_scrollbar_controller_set_zoompoint(plot->dsc_x, 1.0);

   char path[PATH_MAX];

   for(int lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
      if(plot->view[0]->draw.line_set[lsi].dataset->fam_monitored) continue;
      realpath(plot->view[0]->draw.line_set[lsi].dataset->path, path);
      printf("add watch for %s\n", plot->view[0]->draw.line_set[lsi].dataset->path);
      if(FAMMonitorFile(&plot->fc, path, &plot->view[0]->draw.line_set[lsi].dataset->fr, plot)) {
	 fprintf(stderr, "Error cancelling fam monitor for %s: %s\n", plot->view[0]->draw.line_set[lsi].dataset->path, FamErrlist[FAMErrno]);
      }
      plot->view[0]->draw.line_set[lsi].dataset->fam_monitored = 1;
   }
}

static void disable_fam(my_plot_t *plot) {
   if(!plot->fam_initialized) return;

   gtk_widget_set_sensitive(plot->xscroll_off, 1);
   gtk_widget_set_sensitive(plot->xscroll_off_rel, 1);
   gtk_widget_set_sensitive(plot->xscroll_zoompoint, 1);
   gtk_widget_set_sensitive(plot->menuitem_file_refresh, 1);

   for(int lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
      if(!plot->view[0]->draw.line_set[lsi].dataset->fam_monitored) continue;
      printf("remove watch for %s\n", plot->view[0]->draw.line_set[lsi].dataset->path);
      if(FAMCancelMonitor(&plot->fc, &plot->view[0]->draw.line_set[lsi].dataset->fr)) {
	 fprintf(stderr, "Error cancelling fam monitor for %s: %s\n", plot->view[0]->draw.line_set[lsi].dataset->path, FamErrlist[FAMErrno]);
      }
      plot->view[0]->draw.line_set[lsi].dataset->fam_monitored = 0;
   }

   gtk_widget_set_sensitive(plot->menuitem_file_refresh, 1);
}

static void menuitem_file_autorefresh_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean newstate = gtk_check_menu_item_get_active(menuitem);
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(plot->checkbutton_tail), newstate);
   if(newstate) {
      enable_fam(plot);
   } else {
      disable_fam(plot);
   }
}

static void checkbutton_tail_clicked(GtkCheckButton *checkbutton, my_plot_t *plot) {
   gboolean newstate = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton));
   gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(plot->menuitem_file_autorefresh), newstate);
   if(newstate) {
      enable_fam(plot);
   } else {
      disable_fam(plot);
   }
}

#endif

void plot_init(my_plot_t *plot) {
   bzero(plot, sizeof(my_plot_t));

   // init default view
   plot->view_dflt.parent = plot; // back ref
   plot->view_dflt.hline = -1;
   plot->view_dflt.vline = -1;
   plot_draw_init(&plot->view_dflt.draw);
}

void plot_free(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot_draw_deinit(&plot->view[vi]->draw);
      if(plot->view[vi]->draw.font_legend) {
	 free(plot->view[vi]->draw.font_legend);
      }
   }

   data_scrollbar_controller_free(plot->dsc_x);
   for(int vi=0; vi<plot->view_count; ++vi) {
      data_scrollbar_controller_free(plot->view[0]->dsc_y);
   }

   for(int vi=0; vi<plot->view_count; ++vi) {
      if(plot->view[vi]->pangofont) {
	 pango_font_description_free(plot->view[vi]->pangofont);
      }
      if(plot->view[vi]->pangoctx && !plot->view[vi]->da) {
	 g_free(plot->view[vi]->pangoctx);
      }
   }

#ifdef USE_FAM
   if(plot->fam_initialized) {
      g_source_remove(plot->fam_evtsrcid);
      FAMClose(&plot->fc);
   }
#endif

   bzero(plot, sizeof(my_plot_t));
   free(plot);
}

void plot_view_clone(my_plot_view_t *n, my_plot_view_t *o) {
   n->draw.inverse_colors = o->draw.inverse_colors;
   n->draw.samples_per_pixel = o->draw.samples_per_pixel;
   n->draw.show_legend = o->draw.show_legend;
   n->draw.show_disabled = o->draw.show_disabled;

   data_scrollbar_controller_copy(n->dsc_y, o->dsc_y);

   n->draw.xstart = o->draw.xstart;
   n->draw.xend = o->draw.xend;
   n->draw.xmin = o->draw.xmin;
   n->draw.xmax = o->draw.xmax;
   n->draw.ymin = o->draw.ymin;
   n->draw.ymax = o->draw.ymax;

   n->draw.minor_color = o->draw.minor_color;
   n->draw.major_color = o->draw.major_color;
   n->draw.zerogrid_color = o->draw.zerogrid_color;
   n->draw.sep_color = o->draw.sep_color;
   n->draw.label_color = o->draw.label_color;
   n->draw.cursor_marker_color = o->draw.cursor_marker_color;
   n->draw.legend_bg_color = o->draw.legend_bg_color;
   n->draw.legend_border_color = o->draw.legend_border_color;
   n->draw.legend_text_color = o->draw.legend_text_color;

   int setcount = o->draw.line_set_count;
   my_plot_line_info_set_arr_ensure(n->draw.line_set, setcount);

   for(int lsi = 0; lsi < setcount; ++lsi) {
      my_plot_line_info_set_t *ns, *os;
      ns = &n->draw.line_set[lsi];
      os = &o->draw.line_set[lsi];

      ns->dataset = os->dataset;
      ref_dataset(os->dataset);

      int linecount = os->line_count;
      my_plot_line_info_arr_ensure(ns->line, linecount);

      for(int li = 0; li < linecount; ++li) {
	 my_plot_line_info_t *nsl, *osl;
	 nsl = &ns->line[li];
	 osl = &os->line[li];

	 nsl->data = osl->data;
	 nsl->color = osl->color;
	 nsl->dataset_idx = osl->dataset_idx;
	 nsl->enabled = osl->enabled;
	 nsl->mode = osl->mode;
      }
   }
}

my_plot_t *plot_clone(my_plot_t *o) {
   my_plot_t *n = malloc_c(sizeof(my_plot_t));
   plot_init(n);

   my_plot_view_arr_ensure(n->view, o->view_count);

   n->dsc_x = data_scrollbar_controller_clone(o->dsc_x);
   n->fontsize_at_width = o->fontsize_at_width;
   n->scale_font_by_size = o->scale_font_by_size;

   for(int vi=0; vi<o->view_count; ++vi) {
      n->view[vi]->dsc_y = data_scrollbar_controller_clone(o->view[vi]->dsc_y);
      plot_view_clone(n->view[vi], o->view[vi]);
   }

   return n;
}

static char cursor_marker_dashes[] = { 2, 4 };

static my_plot_view_t *plot_get_prio_view(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      if(plot->view[vi]->bmode) {
	 return plot->view[vi];
      }
   }
   return NULL;
}

static void plot_need_redraw(my_plot_t *plot) {
   //printf("need_redraw\n");
   my_plot_view_t *prio = plot_get_prio_view(plot);
   if(prio) {
      // some view is prioritized, render it and keep the others waiting
      plot_view_need_redraw(prio);
      for(int vi=0; vi<plot->view_count; ++vi) {
	 if(plot->view[vi] == prio) continue;
	 plot->view[vi]->normal_expose++;
      }
   } else {
      for(int vi=0; vi<plot->view_count; ++vi) {
	 plot_view_need_redraw(plot->view[vi]);
      }
   }
}

static void plot_view_need_redraw(my_plot_view_t *view) {
   //printf("view_need_redraw\n");
   if(view->conc == 0) {
      if(view->da) gtk_widget_queue_draw(view->da);
   } else {
      view->normal_expose++;
   }
}

void plot_size_or_font_changed(my_plot_t *plot) {
   int w,h;
   for(int vi=0; vi<plot->view_count; ++vi) {
      my_plot_view_t *view = plot->view[vi];

      gdk_drawable_get_size(view->drawable, &w, &h);

      PangoFontDescription *font_desc;
      if(plot->scale_font_by_size) {
	 int fontlen = strlen(plot->font);
	 char *font = malloc((size_t)(fontlen + 40));
	 char *space = strrchr(plot->font, ' ');
	 double size;
	 if(!space) {
	    space = font + fontlen;
	    size = 12;
	 } else {
	    size = atof(space+1);
	 }
	 memcpy(font, plot->font, (size_t)(space - plot->font));

	 // scrollbars etc take 72x110 pixels
	 // good fontsizes:
	 // at 640x480, wh = 530.66 native-fontsize = 9 scaled-fontsize=11
	 // at 640x480, wh = 640 native-fontsize = 10
	 // at 1600x1200, wh = 1491.33 native-fontsize = 17
	 // at 1600x1200, wh = 1600 native-fontsize = 18
	 // first scale according to screen size

	 // scaled = 2.4^(x/1600-1) * 18.3
	 // native = 3.25^(x/1600-1) * 18.3

	 // 18.3 * pow(1.4,sw/1600-1) * pow(2.15, ((pw - 72 + (pw*3/4-110)*4/3)/2)/1600 -1 )

	 double x1 = size;

	 int cur_screen_width = plot->view[0]->da ? gdk_screen_get_width(gtk_window_get_screen(plot->parent)) : 1600;

	 size *= pow(1.4, ((double)cur_screen_width) / 1600 - 1);

	 double x2 = size;

	 // then scale according to widget size
	 double wh = (w + 4.0 * h / 3.0) / 2.0;

	 //size *= exp2(wh / cur_screen_width - 1);
	 //size *= exp2(wh / plot->fontsize_at_width - 1);
	 size *= pow(2.15, wh / 1600 - 1);

	 snprintf(font + (space-plot->font), (size_t)40, " %.1lf", round(size * 2) / 2);

	 //printf("font %s / %lf -> %lf -> %lf   wh %lf  %dx%d   (%d)\n", font, x1, x2, size, wh, w, h, cur_screen_width);

	 font_desc = pango_font_description_from_string(font);

	 free(font);

      } else {
	 font_desc = pango_font_description_from_string(plot->font);
      }

      if(view->pangofont) {
	 pango_font_description_free(view->pangofont);
      }

      view->pangofont = font_desc;

      if(!view->pangoctx) {
	 if(view->da) {
	    view->pangoctx = gtk_widget_get_pango_context(view->da);
	 } else {
	    view->pangoctx = gdk_pango_context_get();
	 }
      }

      // update plot_draw
      if(!view->draw.font_legend) {
	 view->draw.font_legend = zmalloc_c(sizeof(paintapi_font_t));
	 view->draw.font_xaxis = view->draw.font_legend;
	 view->draw.font_yaxis = view->draw.font_legend;
      }
      view->draw.font_legend->ctx = view->pangoctx;
      view->draw.font_legend->font = view->pangofont;

      view->draw.plot_xe = w - 1;
      view->draw.plot_ye = h - 1;
   }

   plot_need_redraw(plot);
}

static void atzh_helper(my_zoom_view_t *z, data_scrollbar_controller_t *dsc) {
   z->off = data_scrollbar_controller_get_page_off(dsc);
   z->length_log = data_scrollbar_controller_get_page_length_log(dsc);
   z->scalepos = data_scrollbar_controller_get_zoompoint(dsc);
}

static void add_to_zoom_history(my_plot_t *plot) {
   if(plot->view[0]->draw.line_set_count == 0) {
      // don't save history if no plots
      return;
   }

   int idx = plot->zoom_pos + 1;
   if(idx > plot->zoom_history_count) idx = plot->zoom_history_count;

   my_zoom_view_arr_ensure(plot->zoom_history, idx + 1);
   atzh_helper(&plot->zoom_history[idx], plot->dsc_x);

   for(int vi=0; vi<plot->view_count; ++vi) {
      atzh_helper(&plot->view[vi]->zoom_history_dflt, plot->view[vi]->dsc_y); // TODO do this in clone instead and get rid of CUSTOM default
      my_zoom_view_arr_ensure(plot->view[vi]->zoom_history, idx + 1);
      atzh_helper(&plot->view[vi]->zoom_history[idx], plot->view[vi]->dsc_y);
   }

   if(idx > 0) {
      gtk_widget_set_sensitive(plot->button_zoom_prev, 1);
   }
   gtk_widget_set_sensitive(plot->button_zoom_next, 0);

   plot->zoom_pos = idx;
}

static void show_all_nohistory(my_plot_view_t *view) {
   double min, max;

   data_scrollbar_controller_get_range(view->parent->dsc_x, &min, &max);
   data_scrollbar_controller_set_page_length(view->parent->dsc_x, max - min);

   data_scrollbar_controller_get_range(view->dsc_y, &min, &max);
   data_scrollbar_controller_set_page_length(view->dsc_y, max - min);
}

static void show_all(my_plot_view_t *view) {
   show_all_nohistory(view);
   add_to_zoom_history(view->parent);
}

static void pan_right(my_plot_t *plot) {
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);

   double min, max;
   data_scrollbar_controller_get_range(plot->dsc_x, &min, &max);

   data_scrollbar_controller_set_page_off(plot->dsc_x, max - xlength);
}

static void switch_to_zoom_history(my_plot_t *plot, int idx) {
   if(idx < 0) return;
   if(idx >= plot->zoom_history_count) return;
   if(idx == plot->zoom_pos) return;

   plot->zoom_pos = idx;

   gtk_widget_set_sensitive(plot->button_zoom_prev, idx > 0);
   gtk_widget_set_sensitive(plot->button_zoom_next, idx < plot->zoom_history_count-1);

   my_zoom_view_t *z = &plot->zoom_history[idx];

   data_scrollbar_controller_set_page_length_log(plot->dsc_x, z->length_log);
   data_scrollbar_controller_set_page_off(plot->dsc_x, z->off);
   data_scrollbar_controller_set_zoompoint(plot->dsc_x, z->scalepos);

   for(int vi=0; vi<plot->view_count; ++vi) {
      z = &plot->view[vi]->zoom_history[idx];

      data_scrollbar_controller_set_page_length_log(plot->view[vi]->dsc_y, z->length_log);
      data_scrollbar_controller_set_page_off(plot->view[vi]->dsc_y, z->off);
      data_scrollbar_controller_set_zoompoint(plot->view[vi]->dsc_y, z->scalepos);
   }
}

static void show_zoom_history(my_plot_t *plot, int idx) {
   if(idx < 0) return;
   if(idx >= plot->zoom_history_count) return;

   my_zoom_view_t *z = &plot->zoom_history[idx];

   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);

   double z__xoff2 = z->off + exp(z->length_log);

   for(int vi=0; vi<plot->view_count; ++vi) {
      my_plot_view_t *view = plot->view[vi];

      z = &view->zoom_history[idx];

      double yoff = data_scrollbar_controller_get_page_off(view->dsc_y);
      double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);

      double z__yoff2 = z->off + exp(z->length_log);

      int x = (int)(view->draw.plot_xe * (z->off - xoff) / xlength);
      int y = (int)(view->draw.plot_yde * (z->off - yoff) / ylength);
      int x2 = (int)(view->draw.plot_xe * (z__xoff2 - xoff) / xlength);
      int y2 = (int)(view->draw.plot_yde * (z__yoff2 - yoff) / ylength);

      y = view->draw.plot_yde - y;
      y2 = view->draw.plot_yde - y2;

      view->bx = x;
      view->by = y;
      view->box = x2;
      view->boy = y2;

      view->draw.paintapi->gc_set_dashes(view->draw.paintapi, view->draw.cursor_marker_gc, -view->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      view->draw.paintapi->draw_rectangle(view->draw.paintapi, view->draw.cursor_marker_gc, 0, x, y, x2, y2);
   }
}

static void hide_zoom_history(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      if(plot->view[vi]->box == COORD_UNDEF) return;

      plot->view[vi]->draw.paintapi->gc_set_dashes(plot->view[vi]->draw.paintapi, plot->view[vi]->draw.cursor_marker_gc, -plot->view[vi]->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      plot->view[vi]->draw.paintapi->draw_rectangle(plot->view[vi]->draw.paintapi, plot->view[vi]->draw.cursor_marker_gc, 0, plot->view[vi]->bx, plot->view[vi]->by, plot->view[vi]->box, plot->view[vi]->boy);

      plot->view[vi]->box = COORD_UNDEF;
   }
}

static void button_zoom_prev_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   switch_to_zoom_history(plot, plot->zoom_pos - 1);
}

static void button_zoom_next_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   switch_to_zoom_history(plot, plot->zoom_pos + 1);
}

static void button_show_all_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   // TODO per-view show-all button ?
   for(int vi=0; vi<plot->view_count; ++vi) {
      show_all_nohistory(plot->view[vi]);
   }
   add_to_zoom_history(plot);
}

static void button_zoom_hover(GtkButton *button, my_plot_t *plot) {
   show_zoom_history(plot, plot->zoom_pos + (GTK_WIDGET(button) == plot->button_zoom_prev ? -1 : 1));
}

static void button_zoom_unhover(GtkButton *button, my_plot_t *plot) {
   hide_zoom_history(plot);
}

static void invert_colors_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean newstate = gtk_check_menu_item_get_active(menuitem);

   if(plot->view[0]->draw.inverse_colors == newstate) return;

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.inverse_colors = newstate;
   }

   reset_gc(plot);

   plot_need_redraw(plot);
}

int plot_add_view_with_xml(my_plot_t *plot, GladeXML *gladexml) {
   int idx = plot->view_count;
   my_plot_view_arr_ensure(plot->view, idx+1);
   my_plot_view_t *view = plot->view[idx];

   view->da = glade_xml_get_widget(gladexml, "drawingarea1");
   view->dsc_y = data_scrollbar_controller_new(gladexml, "yscroll_");
   view->box_yscrolls_plot = glade_xml_get_widget(gladexml, "box_yscrolls_plot");;

   plot_view_connect_signals(view, gladexml);

   return idx;
}

static void plot_add_view(my_plot_t *plot) {
   GladeXML *gladexml = glade_xml_new(plot->gladefile, "box_yscrolls_plot", NULL);
   GtkWidget *box = glade_xml_get_widget(gladexml, "box_yscrolls_plot");
   gtk_box_pack_start(GTK_BOX(plot->box_plots), box, 1, 1, 0);

   int idx = plot_add_view_with_xml(plot, gladexml);

   plot_view_clone(plot->view[idx], plot->view[idx-1]);

   update_view_minmax(plot->view[idx]);
}

static void button_new_view_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   plot_add_view(plot);
}

static void font_response(GtkFontSelectionDialog *dialog, gint response, my_plot_t *plot) {
   if(response == GTK_RESPONSE_OK || response == GTK_RESPONSE_APPLY) {
      if(plot->font) free(plot->font);
      plot->font = g_strdup(gtk_font_selection_dialog_get_font_name(dialog));
      plot->fontsize_at_width = gdk_screen_get_width(gtk_window_get_screen(plot->parent));
      plot_size_or_font_changed(plot);
   }

   if(response != GTK_RESPONSE_APPLY) {
      gtk_widget_destroy(GTK_WIDGET(dialog));
   }
}

static void font_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   GtkFontSelectionDialog *dialog = GTK_FONT_SELECTION_DIALOG(gtk_font_selection_dialog_new("Select font"));
   gtk_widget_show(dialog->apply_button);
   gtk_font_selection_dialog_set_preview_text(dialog, FONT_TEST_STRING);
   gtk_font_selection_dialog_set_font_name(dialog, plot->font);
   /* TODO
   char *font = pango_font_description_to_string(plot->view[0]->pangofont);
   gtk_font_selection_dialog_set_font_name(dialog, font);
   g_free(font);
   */
   g_signal_connect(dialog, "response", G_CALLBACK(font_response), plot);
   gtk_widget_show(GTK_WIDGET(dialog));
}

typedef struct {
      GtkRange *range_sample_density;
      GtkLabel *label_render_time;
      my_plot_t *plot;
} my_plot_sample_density_callback_data_t;

static void sample_density_response(GtkDialog *dialog, gint response, my_plot_sample_density_callback_data_t *sdcbd) {
   if(response == GTK_RESPONSE_OK || response == GTK_RESPONSE_APPLY) {
      double total_render_time = 0;
      double sample_density = exp10(gtk_range_get_value(sdcbd->range_sample_density));
      for(int vi=0; vi<sdcbd->plot->view_count; ++vi) {
	 sdcbd->plot->view[vi]->draw.samples_per_pixel = sample_density;
	 plot_view_need_redraw(sdcbd->plot->view[vi]);
	 gdk_window_process_updates(sdcbd->plot->view[vi]->drawable, 0);

	 total_render_time += sdcbd->plot->view[vi]->draw.last_render_time;
      }
      char tmp[30];
      sprintf(tmp, "<b>%.3lf</b> seconds", total_render_time);
      gtk_label_set_markup(sdcbd->label_render_time, tmp);
   }

   if(response != GTK_RESPONSE_APPLY) {
      bzero(sdcbd, sizeof(my_plot_sample_density_callback_data_t));
      free(sdcbd);
      gtk_widget_destroy(GTK_WIDGET(dialog));
   }
}

static gchar* sample_density_format(GtkScale *scale, gdouble newval, gpointer user_data) {
   if(newval < -0.01) {
      return g_strdup_printf("%.1lf pixels per sample", exp10(-newval));
   } else if(newval <= 0.01) {
      return g_strdup("1 sample per pixel");
   } else {
      return g_strdup_printf("%.1lf samples per pixel", exp10(newval));
   }
}

static void sample_density_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   my_plot_sample_density_callback_data_t *sdcbd = zmalloc_c(sizeof(my_plot_sample_density_callback_data_t));
   sdcbd->plot = plot;

   GladeXML *gladexml = glade_xml_new(plot->gladefile, "dialog_sample_density", NULL);
   GtkWidget *dialog = glade_xml_get_widget(gladexml, "dialog_sample_density");

   sdcbd->range_sample_density = GTK_RANGE(glade_xml_get_widget(gladexml, "hscale_sample_density"));
   gtk_range_set_value(sdcbd->range_sample_density, log10(plot->view[0]->draw.samples_per_pixel));

   sdcbd->label_render_time = GTK_LABEL(glade_xml_get_widget(gladexml, "label_render_time"));
   double total_render_time = 0;
   for(int vi=0; vi<sdcbd->plot->view_count; ++vi) {
      total_render_time += sdcbd->plot->view[vi]->draw.last_render_time;
   }
   char tmp[30];
   sprintf(tmp, "<b>%.3lf</b> seconds", total_render_time);
   gtk_label_set_markup(sdcbd->label_render_time, tmp);

   g_signal_connect(sdcbd->range_sample_density, "format-value", G_CALLBACK(sample_density_format), sdcbd);
   g_signal_connect(dialog, "response", G_CALLBACK(sample_density_response), sdcbd);

   gtk_widget_show(GTK_WIDGET(dialog));
}

static void show_scrollbars_and_tools_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean show = gtk_check_menu_item_get_active(menuitem);
   x_gtk_widget_set_visible(plot->box_xscrolls, show);
   for(int vi=0; vi<plot->view_count; ++vi) {
      x_gtk_widget_set_visible(plot->view[vi]->box_yscrolls, show);
   }
   x_gtk_widget_set_visible(plot->table_tools, show);
}

static void show_disabled_in_legend_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean show = gtk_check_menu_item_get_active(menuitem);
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.show_disabled = show;
   }
   plot_need_redraw(plot);
}

static void show_legend_left_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.show_legend = LEGEND_SHOW_LEFT;
   }
   plot_need_redraw(plot);
}
static void show_legend_right_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.show_legend = LEGEND_SHOW_RIGHT;
   }
   plot_need_redraw(plot);
}
static void hide_legend_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.show_legend = LEGEND_HIDE;
   }
   plot_need_redraw(plot);
}

void show_view_markers(my_plot_view_t *view, boolean_t show) {
   if(view->hmarkers_drawn != show && view->hmarkers_set) {
      view->hmarkers_drawn = show;
      view->draw.paintapi->gc_set_dashes(view->draw.paintapi, view->draw.cursor_marker_gc, -view->vline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      view->draw.paintapi->draw_line(view->draw.paintapi, view->draw.cursor_marker_gc, 0, view->hline, view->draw.plot_xe, view->hline);
   }
   if(view->vmarkers_drawn != show && view->vmarkers_set) {
      view->vmarkers_drawn = show;
      view->draw.paintapi->gc_set_dashes(view->draw.paintapi, view->draw.cursor_marker_gc, -view->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      view->draw.paintapi->draw_line(view->draw.paintapi, view->draw.cursor_marker_gc, view->vline, 0, view->vline, view->draw.plot_yde);
   }

   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 if(view->draw.line_set[lsi].line[li].markers_drawn != show && view->draw.line_set[lsi].line[li].hline != -1) {
	    view->draw.line_set[lsi].line[li].markers_drawn = show;
	    view->draw.paintapi->draw_line(view->draw.paintapi,
					   view->draw.line_set[lsi].line[li].horiz_marker,
					   0, view->draw.line_set[lsi].line[li].hline,
					   view->draw.plot_xe, view->draw.line_set[lsi].line[li].hline);
	 }
      }
   }
}

void show_markers(my_plot_t *plot, boolean_t show) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      show_view_markers(plot->view[vi], show);
   }
}

void reset_view_markers(my_plot_view_t *view) {
   view->hmarkers_drawn = 0;
   view->vmarkers_drawn = 0;
   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 view->draw.line_set[lsi].line[li].markers_drawn = 0;
      }
   }
}

void reset_markers(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      reset_view_markers(plot->view[vi]);
   }
}

void destroy_view_markers(my_plot_view_t *view) {
   show_view_markers(view, 0);
   view->hmarkers_set = 0;
   view->vmarkers_set = 0;
   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 view->draw.line_set[lsi].line[li].hline = -1;
      }
   }
}

void destroy_markers(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      destroy_view_markers(plot->view[vi]);
   }
}

static void set_and_show_markers(my_plot_view_t *view, int x, int y) {
   if(!view->draw.minor_gc) {
      // gui not yet initialized
      return;
   }

   destroy_markers(view->parent);

   if(y > view->draw.plot_yde) return;

   view->draw.paintapi->gc_set_dashes(view->draw.paintapi, view->draw.cursor_marker_gc, -x, cursor_marker_dashes, SZ(cursor_marker_dashes));
   view->draw.paintapi->draw_line(view->draw.paintapi, view->draw.cursor_marker_gc, 0, y, view->draw.plot_xe, y);
   view->hmarkers_drawn = 1;
   view->hmarkers_set = 1;

   int yo = 0;
   for(int vi=0; vi<view->parent->view_count; ++vi) {
      my_plot_view_t *v2 = view->parent->view[vi];
      v2->draw.paintapi->gc_set_dashes(v2->draw.paintapi, v2->draw.cursor_marker_gc, -y, cursor_marker_dashes, SZ(cursor_marker_dashes));
      v2->draw.paintapi->draw_line(v2->draw.paintapi, v2->draw.cursor_marker_gc, x, 0, x, v2->draw.plot_yde);
      v2->hline = y;
      v2->vline = x;
      v2->vmarkers_drawn = 1;
      v2->vmarkers_set = 1;
      yo += v2->draw.plot_ye + 1;
   }

   double xlength = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
   double xoff = data_scrollbar_controller_get_page_off(view->parent->dsc_x);

   double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);
   double yoff = data_scrollbar_controller_get_page_off(view->dsc_y);

   my_time_t xv = (my_time_t)(x * xlength / view->draw.plot_xe + xoff);
   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      int p = my_time_arr_find(view->draw.line_set[lsi].dataset->time, xv, 0);
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 if(!view->draw.line_set[lsi].line[li].enabled) continue;

	 int dataset_idx = view->draw.line_set[lsi].line[li].dataset_idx;

	 int dy = !__isnan(view->draw.line_set[lsi].dataset->item[dataset_idx].data[p]) ?
	    (int)(view->draw.plot_yde - (view->draw.line_set[lsi].dataset->item[dataset_idx].data[p] - yoff) / ylength * view->draw.plot_yde) :
	    -1;

	 if(dy >= 0) {
	    view->draw.paintapi->draw_line(view->draw.paintapi, view->draw.line_set[lsi].line[li].horiz_marker, 0, dy, view->draw.plot_xe, dy);
	    view->draw.line_set[lsi].line[li].hline = dy;
	    view->draw.line_set[lsi].line[li].markers_drawn = 1;
	 }
      }
   }
}

static void plot_move_to(my_plot_view_t *view, int x, int y) {

/*
  data_scrollbar_controller_set_zoompoint(view->dsc_x, ((double)x) / view->draw.plot_xe);
  data_scrollbar_controller_set_zoompoint(view->dsc_y, 1 - ((double)y) / view->draw.plot_yde);
*/

   int xd = x - view->bx;
   int yd = view->by - y;

   xd *= view->bspeed;
   yd *= view->bspeed;

   //printf("move %d,%d\n", xd, yd);

   double xmin, xmax;
   double ymin, ymax;

   data_scrollbar_controller_get_range(view->parent->dsc_x, &xmin, &xmax);
   data_scrollbar_controller_get_range(view->dsc_y, &ymin, &ymax);

   double xlength = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
   double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);

   double nxoff = view->bxoff - xd * xlength / view->draw.plot_xe;
   double nyoff = view->byoff - yd * ylength / view->draw.plot_yde;

   //printf("nt %lf %lf\n", nxoff, nyoff);

   if(nxoff < xmin) nxoff = xmin; else if(nxoff + xlength > xmax) nxoff = xmax - xlength;
   if(nyoff < ymin) nyoff = ymin; else if(nyoff + ylength > ymax) nyoff = ymax - ylength;

   //printf("n %lf %lf\n", nxoff, nyoff);

   data_scrollbar_controller_set_page_off(view->parent->dsc_x, nxoff);
   data_scrollbar_controller_set_page_off(view->dsc_y, nyoff);
}

static int inside_range(int x, int min, int max) {
   return x >= min && x <= max;
}

static int check_if_hit_legend_box(my_plot_view_t *view, int x, int y) {
   return view->draw.show_legend &&
      inside_range(x, view->draw.legend_x, view->draw.legend_x2) &&
      inside_range(y, view->draw.legend_y, view->draw.legend_y2);
}

static my_plot_line_info_t *check_if_hit_legend_item(my_plot_view_t *view, int x, int y) {
   for(int lsi = 0; lsi < view->draw.line_set_count; ++lsi) {
      my_plot_line_info_t *line = view->draw.line_set[lsi].line;
      int line_count = view->draw.line_set[lsi].line_count;

      for(int li=0; li<line_count; ++li) {
	 if(inside_range(x, line[li].legend_x, line[li].legend_x2) &&
	    inside_range(y, line[li].legend_y, line[li].legend_y2)) {
	    return &line[li];
	 }
      }
   }

   return NULL;
}

static gboolean plot_button_press_event(GtkWidget *widget, GdkEventButton *event, my_plot_view_t *view) {
   int x = event->x;
   int y = event->y;

#ifdef USE_FAM
   boolean_t tailing = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(view->parent->checkbutton_tail));
#else
   const boolean_t tailing = 0;
#endif

   switch((int)event->type) {
/*
  case GDK_2BUTTON_PRESS: {

  printf("dpress %d %d\n", x, y);
  plot_move_to(view, view->draw.plot_xe/2, view->draw.plot_yde/2);
  break;
  }
*/

      case GDK_BUTTON_PRESS:
	 // printf("press %d %d\n", x, y);
	 if(view->bmode == 0) {

	    view->bx = x;
	    view->by = y;

	    destroy_markers(view->parent);

	    if(check_if_hit_legend_box(view, x, y)) {
	       my_plot_line_info_t *line = check_if_hit_legend_item(view, x, y);

	       if(line) {
		  switch(event->button) {
		     case 1:
			view->bstate = line->enabled = !line->enabled;
			view->bmode = 4;

			plot_view_need_redraw(view);
			update_view_minmax(view);
			break;

			/*
			  default:
			  diff mode, TODO scales don't work and we need a time unit also
			  view->bstate = line->mode == LINE_MODE_NORMAL;
			  line->mode = view->bstate ? LINE_MODE_DIFF : LINE_MODE_NORMAL;
			  view->bmode = 5;

			  plot_view_need_redraw(view);
			  update_view_minmax(view);
			  break;
			*/
		  }
	       }
	       return 0;
	    }

	    // update scale pos according to press location

	    int snap_x_threshold = 40; // view->draw.plot_xe / 10;
	    int snap_y_threshold = 40; // view->draw.plot_yde / 10;
	    int snap_x = x < snap_x_threshold ? 0 : x >= view->draw.plot_xe - snap_x_threshold ? view->draw.plot_xe-1 : x;
	    int snap_y = y < snap_y_threshold ? 0 : y >= view->draw.plot_yde - snap_y_threshold ? view->draw.plot_yde-1 : y;

	    data_scrollbar_controller_set_zoompoint(view->parent->dsc_x, ((double)snap_x) / view->draw.plot_xe);
	    data_scrollbar_controller_set_zoompoint(view->dsc_y, 1 - ((double)snap_y) / view->draw.plot_yde);

	    view->bspeed = event->state & GDK_SHIFT_MASK ? 4 : 1;

	    if(event->button > 1) {
	       view->bx = x;
	       view->bmode = 2;
	       view->bxloglength = data_scrollbar_controller_get_page_length_log(view->parent->dsc_x);
	       view->byloglength = data_scrollbar_controller_get_page_length_log(view->dsc_y);

	       // when close to border, zoom to border
	       if(tailing || x >= view->draw.plot_xe - 30) {
		  data_scrollbar_controller_set_zoompoint(view->parent->dsc_x, 1.0);
	       } else if(x < 30) { // ##TODO## conf 30
		  data_scrollbar_controller_set_zoompoint(view->parent->dsc_x, 0.0);
	       }

	       if(y < 30) {
		  data_scrollbar_controller_set_zoompoint(view->dsc_y, 1.0);
	       } else if(y >= view->draw.plot_yde - 30) {
		  data_scrollbar_controller_set_zoompoint(view->dsc_y, 0.0);
	       }

	    } else if(event->state & GDK_CONTROL_MASK) {
	       view->bmode = 3;
	       view->box = COORD_UNDEF;

	    } else {
	       view->bx = tailing ? view->draw.plot_xe : x;
	       view->bmode = 1;
	       view->bxoff = data_scrollbar_controller_get_page_off(view->parent->dsc_x);
	       view->byoff = data_scrollbar_controller_get_page_off(view->dsc_y);
	    }
	 }
	 break;
   }

   return 0;
}

#define LIMIT(v,min,max) do{ if(v<(min)) v=min; else if(v>(max)) v=max; } while(0)

static void plot_motion_generic(my_plot_view_t *view, int x, int y, int release) {
#ifdef USE_FAM
   boolean_t tailing = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(view->parent->checkbutton_tail));
#else
   const boolean_t tailing = 0;
#endif

   switch(view->bmode) {
      case 0:
	 if(check_if_hit_legend_box(view, x, y)) {
	    destroy_markers(view->parent);
	 } else {
	    set_and_show_markers(view, x, y);
	 }
	 break;

      case 1:
	 plot_move_to(view, tailing ? view->draw.plot_xe : x, y);
	 if(release) {
	    add_to_zoom_history(view->parent);
	 }
	 break;

      case 2: {
	 // ##TODO## conf 5.0 and 2.0
	 data_scrollbar_controller_set_page_length_log(view->parent->dsc_x, view->bxloglength - (x - view->bx) * view->bspeed * 5.0 / view->draw.plot_xe);
	 data_scrollbar_controller_set_page_length_log(view->dsc_y, view->byloglength - (y - view->by) * view->bspeed * 2.0 / view->draw.plot_yde);
	 if(release) {
	    add_to_zoom_history(view->parent);
	 }
	 break;
      }

      case 3: {
	 LIMIT(view->bx, 0, view->draw.plot_xe);
	 LIMIT(view->by, 0, view->draw.plot_yde);
	 LIMIT(x, 0, view->draw.plot_xe);
	 LIMIT(y, 0, view->draw.plot_yde);

	 view->draw.paintapi->gc_set_dashes(view->draw.paintapi, view->draw.cursor_marker_gc, -view->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
	 if(!release) {
	    view->draw.paintapi->draw_rectangle(view->draw.paintapi, view->draw.cursor_marker_gc, 0, view->bx, view->by, x, y);
	 }
	 if(view->box != COORD_UNDEF) {
	    view->draw.paintapi->draw_rectangle(view->draw.paintapi, view->draw.cursor_marker_gc, 0, view->bx, view->by, view->box, view->boy);
	 }
	 if(!release) {
	    view->box = x;
	    view->boy = y;
	 } else {
	    int x1 = view->bx;
	    int y1 = view->by;

	    y = view->draw.plot_yde - y;
	    y1 = view->draw.plot_yde - y1;

	    if(x < x1) { int t = x; x = x1; x1 = t; }
	    if(y < y1) { int t = y; y = y1; y1 = t; }

	    double ox = data_scrollbar_controller_get_page_off(view->parent->dsc_x);
	    double oy = data_scrollbar_controller_get_page_off(view->dsc_y);
	    double lx = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
	    double ly = data_scrollbar_controller_get_page_length(view->dsc_y);

	    //printf("%d,%d %d,%d\n", x1, y1, x-x1, y-y1);

	    //printf("%lf,%lf %lf,%lf\n", ox, oy, lx, ly);

	    ox += lx * x1 / view->draw.plot_xe;
	    oy += ly * y1 / view->draw.plot_yde;

	    lx *= (x - x1) / (double)view->draw.plot_xe;
	    ly *= (y - y1) / (double)view->draw.plot_yde;

	    //printf("%lf,%lf %lf,%lf\n\n", ox, oy, lx, ly);

	    // after area zoom, always put center as zoom point
	    data_scrollbar_controller_set_zoompoint(view->parent->dsc_x, 0.5);
	    data_scrollbar_controller_set_zoompoint(view->dsc_y, 0.5);

	    // update page lengths
	    data_scrollbar_controller_set_page_length(view->parent->dsc_x, lx);
	    data_scrollbar_controller_set_page_length(view->dsc_y, ly);

	    // now, in case either page length subceeded the minimum page length, we adjust offset so that the selected zoom area becomes centered in view
	    double nlx = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
	    double nly = data_scrollbar_controller_get_page_length(view->dsc_y);

	    ox -= (nlx - lx) / 2;
	    oy -= (nly - ly) / 2;

	    // update offsets
	    data_scrollbar_controller_set_page_off(view->parent->dsc_x, ox);
	    data_scrollbar_controller_set_page_off(view->dsc_y, oy);

	    //printf("%lf,%lf %lf,%lf\n\n", data_scrollbar_controller_get_page_off(view->parent->dsc_x), data_scrollbar_controller_get_page_off(view->dsc_y), nlx, nly);

	    view->box = COORD_UNDEF;

	    add_to_zoom_history(view->parent);
	 }
	 break;
      }

      case 4: {
	 if(!view->draw.show_disabled) break;
	 if(!check_if_hit_legend_box(view, x, y)) break;

	 my_plot_line_info_t *line = check_if_hit_legend_item(view, x, y);

	 if(line) {
	    int changed = line->enabled != view->bstate;
	    if(changed) {
	       line->enabled = view->bstate;

	       plot_view_need_redraw(view);
	       update_view_minmax(view);
	    }
	 }
	 break;
      }

      case 5: {
	 if(!check_if_hit_legend_box(view, x, y)) break;

	 my_plot_line_info_t *line = check_if_hit_legend_item(view, x, y);

	 if(line) {
	    int changed = line->mode != (view->bstate ? LINE_MODE_DIFF : LINE_MODE_NORMAL);
	    if(changed) {
	       line->mode = view->bstate ? LINE_MODE_DIFF : LINE_MODE_NORMAL;

	       plot_view_need_redraw(view);
	       update_view_minmax(view);
	    }
	 }
	 break;
      }

      default:
	 fprintf(stderr, "ERROR: bmode = %d\n", view->bmode);
	 break;
   }
}

static gboolean plot_button_release_event(GtkWidget *widget, GdkEventButton *event, my_plot_view_t *view) {
   int x = event->x;
   int y = event->y;

   plot_motion_generic(view, x, y, 1);

   view->bmode = 0;

   return 0;
}

static gboolean plot_leave_notify_event(GtkWidget *widget, GdkEventCrossing *event, my_plot_view_t *view) {
   destroy_markers(view->parent);

   return 0;
}

static gboolean plot_motion_notify_event(GtkWidget *wid, GdkEventMotion *event, my_plot_view_t *view) {
   int x = event->x;
   int y = event->y;

   plot_motion_generic(view, x, y, 0);

   return 0;
}

/*
  my $co_minor = $#colors - 3;
  my $co_major = $#colors - 2;
  my $co_sep = $#colors - 1;
  my $co_label = $#colors - 0;
*/

static void plot_generic_scrollbar_value_changed(my_plot_t *plot) {
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double ylength = data_scrollbar_controller_get_page_length(plot->view[0]->dsc_y);

   double xlo, xhi;
   double ylo, yhi;
   data_scrollbar_controller_get_range(plot->dsc_x, &xlo, &xhi);
   data_scrollbar_controller_get_range(plot->view[0]->dsc_y, &ylo, &yhi); // TODO multiview

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.xstart = (my_time_t)xlo;
      plot->view[vi]->draw.xend = (my_time_t)xhi;
   }

   gboolean show_all_enable =
      (xhi-xlo - xlength) / (xhi - xlo) > 1e-10 ||
      (yhi-ylo - ylength) / (yhi - ylo) > 1e-10;

   if(plot->button_show_all) {
      gtk_widget_set_sensitive(plot->button_show_all, show_all_enable);
   }
   if(plot->menuitem_show_all) {
      gtk_widget_set_sensitive(plot->menuitem_show_all, show_all_enable);
   }
}

static void plot_xscrollbar_value_changed(GtkRange *not_always_provided, my_plot_t *plot) {

   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.xmin = xoff;
      plot->view[vi]->draw.xmax = xoff + xlength;
   }

   plot_generic_scrollbar_value_changed(plot);

   plot_need_redraw(plot);
}

static void plot_yscrollbar_value_changed(GtkRange *not_always_provided, my_plot_view_t *view) {

   double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);
   double yoff = data_scrollbar_controller_get_page_off(view->dsc_y);

   view->draw.ymin = yoff;
   view->draw.ymax = yoff + ylength;

   plot_generic_scrollbar_value_changed(view->parent);

   plot_view_need_redraw(view);
}

static gboolean plot_configure_event(GtkWidget *widget, GdkEventConfigure *do_not_use, my_plot_view_t *view) {
   plot_size_or_font_changed(view->parent);
   return 0;
}

static gboolean plot_map_event(GtkWidget *widget, GdkEvent *do_not_use, my_plot_view_t *view) {
   //printf("  --  map\n");

   view->drawable = view->da->window;

   view->draw.paintapi = paintapi_gtk_new(view->drawable, &view->draw);

   // set the background part 1 - create new style for widget
   GtkStyle *style = gtk_widget_get_style(view->da);
   style = gtk_style_copy(style);
   gtk_widget_set_style(view->da, style);

   // set the background part 2
   GdkColor bg;
   get_color(view->parent, &bg, "#000000");
   gdk_window_set_background(view->drawable, &bg);

   // set cursor
   GdkCursor *cursor = gdk_cursor_new(GDK_DRAFT_SMALL);
   gdk_window_set_cursor(view->da->window, cursor);
   gdk_cursor_unref(cursor);

   plot_configure_event(widget, NULL, view);

   return 0;
}

//static int indent = 0;

static gboolean plot_expose_event(GtkWidget *widget, GdkEventExpose *event, my_plot_view_t *view) {

   if(view->conc > 0) {
      // still not optimal
      view->normal_expose++;
      //printf(" -> normal_expose++\n");
      return 0;
   }
   int mytrans = ++view->parent->dsc_x->num_changes + view->dsc_y->num_changes;
   view->conc++;
   //    print "->$mytrans view->conc\n";

   //int mvi; for(mvi=0; mvi<view->parent->view_count; ++mvi) if(view->parent->view[mvi] == view) break;
   //printf("%*sexpose %d\n", indent++*3, "", mvi);

   view->box = COORD_UNDEF;

   reset_view_markers(view);

   plot_draw(&view->draw);

   show_markers(view->parent, 1);

   if(mytrans != view->parent->dsc_x->num_changes + view->dsc_y->num_changes || view->normal_expose) {
      //printf("%d pending exposes, queue draw\n", view->normal_expose);
      //printf("%*sRedraw self (%d)\n", indent*3, "", mvi);
      gtk_widget_queue_draw(view->da);
   } else {
      for(int vi=0; vi<view->parent->view_count; ++vi) {
	 if(view->parent->view[vi]->normal_expose) {
	    //printf("%*sRedraw %d\n", indent*3, "", vi);
	    plot_view_need_redraw(view->parent->view[vi]);
	    break;
	 }
      }
   }
   view->normal_expose = 0;
   view->conc--;

   //printf("%*s/expose %d\n", --indent*3, "", mvi);
   
   return 0;
}

// TODO remove this
int get_black_color(GdkColor *dst) {
   GdkColormap *colormap = gdk_colormap_get_system();
   if(!gdk_color_parse("black", dst)) return 1;
   return gdk_colormap_alloc_color(colormap, dst, 0, 0) ? 0 : 2;
}

int get_color(my_plot_t *plot, GdkColor *dst, const char *str) {
   GdkColormap *colormap = gdk_colormap_get_system();
   if(!gdk_color_parse(plot->view[0]->draw.inverse_colors ? "white" : "black", dst)) return 1; // it's a hack!
   return gdk_colormap_alloc_color(colormap, dst, 0, 0) ? 0 : 2;
}

// TODO remove this
GdkGC *x_gdk_gc_new_with_fg_color(my_plot_t *plot, const char *color) {
   GdkGC *gc = gdk_gc_new(plot->view[0]->drawable); // TODO hack, uses same gc for multiple drawable, does it even work ?
   // gdk_gc_copy(line[li].line_gc, plot->da->style->fg_gc[GTK_STATE_NORMAL]);
   GdkColor c;
   get_color(plot, &c, color);
   gdk_gc_set_foreground(gc, &c);
   return gc;
}

#define VDIFFMIN_DIVISOR 1000

void update_view_minmax(my_plot_view_t *view) {
   double vmin = INFINITY, vmax = -INFINITY;
   double vdiffmin = VDIFFMIN_DIVISOR;

   int count = 0;

   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      my_dataset_t *dataset = view->draw.line_set[lsi].dataset;
      my_plot_line_info_t *line = view->draw.line_set[lsi].line;
      int line_count = view->draw.line_set[lsi].line_count;

      for(int li=0; li<line_count; ++li) {
	 if(!line[li].enabled) continue;
	 int d = line[li].dataset_idx;
	 switch(line[li].mode) {
	    case LINE_MODE_NORMAL: {
	       if(count == 0 || vmin > dataset->item[d].min) vmin = dataset->item[d].min;
	       if(count == 0 || vmax < dataset->item[d].max) vmax = dataset->item[d].max;

	       // TODO should we include disabled lines in vdiffmin calculation...
	       double vdiff = dataset->item[d].max - dataset->item[d].min;
	       if(count == 0 || vdiffmin > vdiff) vdiffmin = vdiff;

	       break;
	    }
	    case LINE_MODE_DIFF: {
	       if(count == 0 || vmin > dataset->item[d].min_diff) vmin = dataset->item[d].min_diff;
	       if(count == 0 || vmax < dataset->item[d].max_diff) vmax = dataset->item[d].max_diff;

	       // TODO see above
	       double vdiff = dataset->item[d].max_diff - dataset->item[d].min_diff;
	       if(count == 0 || vdiffmin > vdiff) vdiffmin = vdiff;

	       break;
	    }
	 }
	 count++;
      }
   }

   // don't do anything if no lines visible
   if(count == 0) return;

   data_scrollbar_controller_set_range(view->dsc_y, vmin - (vmax-vmin)/20, vmax + (vmax-vmin)/20);

   //data_scrollbar_controller_set_page_length(plot->dsc_y, vmax-vmin); // initially zoom out

   // TODO maybe use diff data as base for minimum length instead of a fraction of the full range

   double yminlength = vdiffmin > 0 ? vdiffmin / VDIFFMIN_DIVISOR : 1;
   data_scrollbar_controller_set_min_interval(view->dsc_y, yminlength);
}

void update_x_minmax(my_plot_t *plot) {
   my_time_t ts = 0, te = 1;
   my_time_t min_time_diff = 1;

   int lsi;
   for(lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
      my_dataset_t *dataset = plot->view[0]->draw.line_set[lsi].dataset;

      if(lsi == 0 || min_time_diff > dataset->min_time_diff) {
	 min_time_diff = dataset->min_time_diff;
      }

      if(lsi == 0 || dataset->ts < ts) {
	 ts = dataset->ts;
      }

      if(lsi == 0 || dataset->te > te) {
	 te = dataset->te;
      }
   }

   // don't do anything if no datasets available
   if(lsi == 0) return;

   data_scrollbar_controller_set_range(plot->dsc_x, (double)ts, (double)te);
   data_scrollbar_controller_set_min_interval(plot->dsc_x, min_time_diff * 10.0);
}

void update_minmax(my_plot_t *plot) {
   update_x_minmax(plot);

   for(int vi=0; vi<plot->view_count; ++vi) {
      update_view_minmax(plot->view[vi]);
   }
}

int linecoloroff = 0;

static void update_progress(double percent, void *plotp) {
   my_plot_t *plot = (my_plot_t*)plotp;

   gtk_progress_bar_set_fraction(plot->progressbar, percent);
}

void add_datafile(my_plot_t *plot, char *filename) {
   printf("Loading %s ..\n", filename);

   char tmp[40 + strlen(filename)];
   sprintf(tmp, "Loading %s ..", filename);
   gtk_widget_show(GTK_WIDGET(plot->progressbar));
   gtk_progress_bar_set_text(plot->progressbar, tmp);

   gtk_widget_set_sensitive(plot->menuitem_file_add_data_file, 0);

   my_dataset_t *dataset = read_dataset(filename, update_progress, plot);

   gtk_widget_set_sensitive(plot->menuitem_file_add_data_file, 1);

   gtk_progress_bar_set_text(plot->progressbar, "");
   gtk_progress_bar_set_fraction(plot->progressbar, 0.0);
   gtk_widget_hide(GTK_WIDGET(plot->progressbar));

   if(!dataset) {
      if(!quit) {
	 fprintf(stderr, "\033[31;1mWARNING: File '%s' could not be loaded\033[m\n", filename);
      }
      return;
   }

   dataset->maxdiff = 100 * TIME_MULTIPLIER;

   int lsi = plot->view[0]->draw.line_set_count;
   for(int vi=0; vi<plot->view_count; ++vi) {
      my_plot_view_t *view = plot->view[vi];

      my_plot_line_info_set_arr_ensure(view->draw.line_set, lsi + 1);

      view->draw.line_set[lsi].dataset = dataset;

      if(vi == 0) printf("Datafile %d: %d data columns, %d samples\n", lsi + 1, dataset->item_count, dataset->time_count);

      for(int dsi=0; dsi<dataset->item_count; ++dsi) {
	 int li = view->draw.line_set[lsi].line_count;
	 my_plot_line_info_arr_ensure(view->draw.line_set[lsi].line, li+1);
	 int linecoloridx = linecoloroff + dsi;
	 view->draw.line_set[lsi].line[li].color = default_line_colors[linecoloridx % SZ(default_line_colors)];
	 view->draw.line_set[lsi].line[li].dataset_idx = dsi;
	 view->draw.line_set[lsi].line[li].enabled = 1;
	 view->draw.line_set[lsi].line[li].mode = LINE_MODE_NORMAL;
	 if(vi == 0) printf("  column %d: value range %lf - %lf\n", dsi+1, dataset->item[dsi].min, dataset->item[dsi].max);
      }
   }

   linecoloroff += dataset->item_count;

   puts("");

#ifdef USE_FAM
   boolean_t tailing = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(plot->checkbutton_tail));
   if(tailing) {
      enable_fam(plot);
   }
#endif

   plot_need_redraw(plot);
   update_minmax(plot);
   for(int vi=0; vi<plot->view_count; ++vi) {
      show_all_nohistory(plot->view[vi]);
   }

   if(lsi == 0) {
      // set initial center
      double min, max;
      for(int vi=0; vi<plot->view_count; ++vi) {
	 data_scrollbar_controller_get_range(plot->view[vi]->dsc_y, &min, &max);
	 data_scrollbar_controller_set_zoompoint(plot->view[vi]->dsc_y, min > 0 ? 0 : max < 0 ? 1 : -min / (max - min)); // align to(wards) zero
      }
      data_scrollbar_controller_set_zoompoint(plot->dsc_x, 1.0); // align to right
   }

   add_to_zoom_history(plot);
}

static void screen_size_changed(GdkScreen *screen, my_plot_t *plot) {
   // works only with xrandr, not with ctrl-alt-+ and ctrl-alt--
   plot_size_or_font_changed(plot);
}

typedef enum { DND_INFO_URI, DND_INFO_TEXT } my_drag_drop_info_t;

static void drag_n_drop(GtkWidget *widget, GdkDragContext *context, gint x, gint y, GtkSelectionData *data, guint info, guint drag_time, my_plot_t *plot) {
   switch(info) {
      case DND_INFO_URI: {
	 gchar **uris = gtk_selection_data_get_uris(data);
	 if(!uris) break;
	 for(int i=0; uris[i]; ++i) {
	    printf("Got %s\n", uris[i]);
	    char *hostname;
	    char *path = g_filename_from_uri(uris[i], &hostname, NULL);
	    if(path && (!hostname || !strncmp(hostname, "localhost", (size_t)9))) {
	       add_datafile(plot, path);
	    }
	    if(path) free(path);
	    if(hostname) free(hostname);
	 }
	 g_strfreev (uris);
	 break;
      }
      case DND_INFO_TEXT: {
	 gchar *text = (gchar *)gtk_selection_data_get_text(data);
	 if(!text) break;
	 // assume text is a set of lines, each line containing a filename
	 gchar **paths = g_strsplit_set(text, "\r\n", 0);
	 for(int i=0; paths[i]; ++i) {
	    if(strspn(paths[i], " \t") == strlen(paths[i])) continue; // only whitespace, if any
	    add_datafile(plot, paths[i]);
	 }
	 g_strfreev(paths);
	 free(text);
	 break;
      }
   }
}

static void plot_view_connect_signals(my_plot_view_t *view, GladeXML *gladexml) {

   // setup drag and drop

   gtk_drag_dest_set(view->da, GTK_DEST_DEFAULT_ALL, NULL, 0, GDK_ACTION_COPY);
   GtkTargetList *target_list = gtk_target_list_new(NULL, 0);
   gtk_target_list_add_uri_targets(target_list, DND_INFO_URI);
   gtk_target_list_add_text_targets(target_list, DND_INFO_TEXT);
   gtk_drag_dest_set_target_list(view->da, target_list);
   gtk_target_list_unref(target_list);
   g_signal_connect(view->da, "drag-data-received", G_CALLBACK(drag_n_drop), view);

   // drawing area

   gtk_widget_add_events(view->da, GDK_LEAVE_NOTIFY_MASK);
   g_signal_connect(view->da, "leave-notify-event", G_CALLBACK(plot_leave_notify_event), view);

   gtk_widget_add_events(view->da, GDK_POINTER_MOTION_MASK);
   g_signal_connect(view->da, "motion-notify-event", G_CALLBACK(plot_motion_notify_event), view);

   //gtk_widget_add_events(view->da, GDK_BUTTON_MOTION_MASK);
   //g_signal_connect(view->da, "-event", G_CALLBACK(plot_motion_notify_event), view);

   gtk_widget_add_events(view->da, GDK_BUTTON_PRESS_MASK);
   g_signal_connect(view->da, "button-press-event", G_CALLBACK(plot_button_press_event), view);

   gtk_widget_add_events(view->da, GDK_BUTTON_RELEASE_MASK);
   g_signal_connect(view->da, "button-release-event", G_CALLBACK(plot_button_release_event), view);


   g_signal_connect(view->da, "map-event", G_CALLBACK(plot_map_event), view);
   g_signal_connect(view->da, "expose-event", G_CALLBACK(plot_expose_event), view);
   g_signal_connect(view->da, "configure-event", G_CALLBACK(plot_configure_event), view);

   // y scrollbars

   g_signal_connect(gtk_range_get_adjustment(GTK_RANGE(glade_xml_get_widget(gladexml, "yscroll_scrollbar_off"))), "value-changed", G_CALLBACK(plot_yscrollbar_value_changed), view);
   g_signal_connect(gtk_range_get_adjustment(GTK_RANGE(glade_xml_get_widget(gladexml, "yscroll_scrollbar_off"))), "changed", G_CALLBACK(plot_yscrollbar_value_changed), view);

   view->box_yscrolls = glade_xml_get_widget(gladexml, "box_yscrolls");

   plot_yscrollbar_value_changed(NULL, view);
}

void plot_connect_signals(my_plot_t *plot, const char *gladefile, GladeXML *gladexml) {

  GdkScreen* screen = gtk_window_get_screen(plot->parent);
  g_signal_connect(screen, "size-changed", G_CALLBACK(screen_size_changed), plot);

/*
  GdkRectangle r;
  gdk_screen_get_monitor_geometry(screen, 0, &r);

  printf("screen %d,%d\n", gdk_screen_get_width(screen), gdk_screen_get_height(screen));
  printf("  mm   %d,%d\n", gdk_screen_get_width_mm(screen), gdk_screen_get_height_mm(screen));
  printf("  moni %d,%d\n", r.width, r.height);
*/

   // buttons

   plot->button_show_all = glade_xml_get_widget(gladexml, "button_show_all");
   g_signal_connect(plot->button_show_all, "clicked", G_CALLBACK(button_show_all_clicked), plot);

   g_signal_connect(plot->button_zoom_prev, "clicked", G_CALLBACK(button_zoom_prev_clicked), plot);
   g_signal_connect(plot->button_zoom_next, "clicked", G_CALLBACK(button_zoom_next_clicked), plot);
   g_signal_connect(plot->button_zoom_prev, "enter", G_CALLBACK(button_zoom_hover), plot);
   g_signal_connect(plot->button_zoom_next, "enter", G_CALLBACK(button_zoom_hover), plot);
   g_signal_connect(plot->button_zoom_prev, "leave", G_CALLBACK(button_zoom_unhover), plot);
   g_signal_connect(plot->button_zoom_next, "leave", G_CALLBACK(button_zoom_unhover), plot);

   plot->box_plots = glade_xml_get_widget(gladexml, "vbox_window");
   g_signal_connect(glade_xml_get_widget(gladexml, "button_new_view"), "clicked", G_CALLBACK(button_new_view_clicked), plot);

   //g_signal_connect(glade_xml_get_widget(gladexml, "button_menu"), "clicked", G_CALLBACK(menu_clicked), glade_xml_get_widget(gladexml, "menubar_main"));

   // menu

   plot->box_xscrolls = glade_xml_get_widget(gladexml, "box_xscrolls");
   plot->table_tools = glade_xml_get_widget(gladexml, "table_tools");

#ifdef USE_FAM

   plot->menuitem_file_refresh = glade_xml_get_widget(gladexml, "menuitem_file_refresh");
   plot->menuitem_file_autorefresh = glade_xml_get_widget(gladexml, "menuitem_file_autorefresh");
   plot->checkbutton_tail = glade_xml_get_widget(gladexml, "checkbutton_tail");

   plot->xscroll_off = glade_xml_get_widget(gladexml, "xscroll_scrollbar_off");
   plot->xscroll_off_rel = glade_xml_get_widget(gladexml, "xscroll_scrollbar_off_rel");
   plot->xscroll_zoompoint = glade_xml_get_widget(gladexml, "xscroll_scale");

   gtk_widget_set_sensitive(plot->menuitem_file_autorefresh, 1);
   gtk_widget_set_sensitive(plot->checkbutton_tail, 1);

   g_signal_connect(plot->menuitem_file_autorefresh, "activate", G_CALLBACK(menuitem_file_autorefresh_clicked), plot);
   g_signal_connect(plot->checkbutton_tail, "toggled", G_CALLBACK(checkbutton_tail_clicked), plot);

#endif

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_add_data_file"), "activate", G_CALLBACK(add_data_file_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_remove_data_file"), "activate", G_CALLBACK(remove_data_file_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_remove_all_data_files"), "activate", G_CALLBACK(remove_all_data_files_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_refresh"), "activate", G_CALLBACK(refresh_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_export_as"), "activate", G_CALLBACK(export_as_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_quit"), "activate", G_CALLBACK(request_quit), (gpointer)2);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_view_previous"), "activate", G_CALLBACK(button_zoom_prev_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_view_next"), "activate", G_CALLBACK(button_zoom_next_clicked), plot);

   plot->menuitem_show_all = glade_xml_get_widget(gladexml, "menuitem_show_all");
   g_signal_connect(plot->menuitem_show_all, "activate", G_CALLBACK(button_show_all_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_new_view"), "activate", G_CALLBACK(button_new_view_clicked), plot);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_inverse_colors"), "activate", G_CALLBACK(invert_colors_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_font"), "activate", G_CALLBACK(font_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_sample_density"), "activate", G_CALLBACK(sample_density_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_legend_left"), "activate", G_CALLBACK(show_legend_left_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_legend_right"), "activate", G_CALLBACK(show_legend_right_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_hide_legend"), "activate", G_CALLBACK(hide_legend_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_scrollbars_and_tools"), "activate", G_CALLBACK(show_scrollbars_and_tools_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_disabled_in_legend"), "activate", G_CALLBACK(show_disabled_in_legend_clicked), plot);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_about"), "activate", G_CALLBACK(about_clicked), plot);

   // x scrollbars

   g_signal_connect(gtk_range_get_adjustment(GTK_RANGE(glade_xml_get_widget(gladexml, "xscroll_scrollbar_off"))), "value-changed", G_CALLBACK(plot_xscrollbar_value_changed), plot);
   g_signal_connect(gtk_range_get_adjustment(GTK_RANGE(glade_xml_get_widget(gladexml, "xscroll_scrollbar_off"))), "changed", G_CALLBACK(plot_xscrollbar_value_changed), plot);

   plot_xscrollbar_value_changed(NULL, plot);
}
