/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * plot_draw - handles the drawing of a plot
 */

#include <math.h>
#include <sys/time.h>

#include "plot_draw.xh"
#include "util.xh"
#include "grid.xh"
#include "main.xh"

#if 0
#define pprintf printf
#else
#define pprintf(...)
#endif

static char major_dashes[] = { 2, 2 };
static char minor_dashes[] = { 1, 3 };

static void *gc_new_with_color(my_plot_draw_t *plot, paintapi_rgb_t color) {
   void *gc = plot->paintapi->gc_new(plot->paintapi);
   plot->paintapi->gc_set_foreground(plot->paintapi, gc, &color);
   return gc;
}

void plot_setup_gcs(my_plot_draw_t *plot) {
   // ##TODO##
   int lsi;
   for(lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;
      int i;
      for(i=0; i<line_count; ++i) {
	 if(likely(line[i].line_gc)) continue;
	 // printf("%d %s\n", i, line[i].color);

	 line[i].line_gc = gc_new_with_color(plot, line[i].color);
	 line[i].horiz_marker = gc_new_with_color(plot, line[i].color);

	 char horiz_marker_dashes[] = { 2, 4 };
	 plot->paintapi->gc_set_dashes(plot->paintapi, line[i].horiz_marker, 0, horiz_marker_dashes, SZ(horiz_marker_dashes));
	 plot->paintapi->gc_set_function(plot->paintapi, line[i].horiz_marker, PAINTAPI_FUNCTION_XOR);
      }
   }

   if(unlikely(!plot->minor_gc)) {
      plot->minor_gc = gc_new_with_color(plot, plot->minor_color);
      plot->major_gc = gc_new_with_color(plot, plot->major_color);
      plot->zerogrid_gc = gc_new_with_color(plot, plot->zerogrid_color);
      plot->sep_gc = gc_new_with_color(plot, plot->sep_color);
      plot->label_gc = gc_new_with_color(plot, plot->label_color);
      plot->cursor_marker_gc = gc_new_with_color(plot, plot->cursor_marker_color);

      plot->paintapi->gc_set_function(plot->paintapi, plot->cursor_marker_gc, PAINTAPI_FUNCTION_XOR);

      //	foreach my $family ($da->get_pango_context()->get_font_map()->list_families()) {
      //	    print %{$family}."\n";
      //	}

      plot->legend_bg_gc = gc_new_with_color(plot, plot->legend_bg_color);
      plot->legend_border_gc = gc_new_with_color(plot, plot->legend_border_color);
      plot->legend_text_gc = gc_new_with_color(plot, plot->legend_text_color);
   }
}

void plot_draw_value_grid(my_plot_draw_t *plot) {

   pprintf("y %lf %lf\n", plot->ymin, plot->ymax);

   grid_t hgrid = calc_value_grid(plot->ymin, plot->ymax, plot->plot_yde / (20 + 2 * plot->font_height), plot->plot_yde + 1);

   pprintf("%lf %lf %lf %lf %lf %lf %d\n", hgrid.dmin, hgrid.d, hgrid.dmax, hgrid.dvmin, hgrid.dv, hgrid.dvmax, hgrid.minpermaj);

   double xlength = plot->xmax - plot->xmin;
   int doff = (int)(plot->xmin * plot->plot_xe / xlength);
   plot->paintapi->gc_set_dashes(plot->paintapi, plot->minor_gc, doff, minor_dashes, SZ(minor_dashes));
   plot->paintapi->gc_set_dashes(plot->paintapi, plot->major_gc, doff, major_dashes, SZ(major_dashes));
   plot->paintapi->gc_set_dashes(plot->paintapi, plot->zerogrid_gc, doff, major_dashes, SZ(major_dashes));

   // how many decimals to display
   int decimals = (int)-floor(log10(hgrid.dv) + 0.05); // 0.05 is for rounding erros
   if(decimals < 0) decimals = 0;

   plot->value_axis_width = 0;

   double yo,yvo;
   for(yo = hgrid.dmin, yvo = hgrid.dvmin; yo < hgrid.dmax + hgrid.d; yo += hgrid.d, yvo += hgrid.dv) {
      if(yo < 0) continue;
      int y = plot->plot_yde - (int)yo;
      char tmp[30];
      snprintf(tmp, sizeof(tmp), "%.*lf", decimals, yvo);

      paintapi_gc_t *gc;
      if(fabs(yvo / hgrid.dv) < 1e-10) {
	 gc = plot->zerogrid_gc;
	 snprintf(tmp, sizeof(tmp), "%.*lf", decimals, 0.0); // to avoid -0.0
      } else if(((int)(round(yvo / hgrid.dv))) % hgrid.minpermaj == 0) {
	 gc = plot->major_gc;
      } else {
	 gc = plot->minor_gc;
      }

      paintapi_textlayout_t *layout = plot->paintapi->textlayout_create(plot->paintapi, plot->font_yaxis, tmp);

      paintapi_textlayout_extents_t ri;
      plot->paintapi->textlayout_calculate_size(plot->paintapi, layout, &ri);
      plot->value_axis_width = max(plot->value_axis_width, ri.xr - ri.xl + 1);

      plot->paintapi->draw_textlayout(plot->paintapi, plot->label_gc, 0, y + 2 - ri.yt, layout);
      plot->paintapi->draw_line(plot->paintapi, gc, 0, y, plot->plot_xe, y);

      plot->paintapi->textlayout_free(plot->paintapi, layout);
   }
}

void plot_draw_time_grid(my_plot_draw_t *plot) {

   pprintf("x %lf %lf\n", plot->xmin, plot->xmax);

   grid_t vgrid;

   time_field_t max_diff_field = time_diff_field(plot->xend - plot->xstart);
   time_field_t min_diff_field = TIME_FIELD_MILLISEC;

   if(max_diff_field == TIME_FIELD_MILLISEC) max_diff_field = TIME_FIELD_SEC;
   if(max_diff_field == TIME_FIELD_MONTH) max_diff_field = TIME_FIELD_YEAR;

   paintapi_textlayout_extents_t ri;

   double min = plot->xmin;
   for(; min_diff_field < max_diff_field; min_diff_field++) {
      if(min_diff_field == TIME_FIELD_HOUR) min_diff_field = TIME_FIELD_DAY;

      //printf("%d %d\n", min_diff_field, max_diff_field);

      const char *example_time = format_time2((my_time_t)round(min), min_diff_field, max_diff_field);
      //printf("-- %s --\n", example_time);
      paintapi_textlayout_t *layout = plot->paintapi->textlayout_create(plot->paintapi, plot->font_xaxis, example_time);
      plot->paintapi->textlayout_calculate_size(plot->paintapi, layout, &ri);
      plot->paintapi->textlayout_free(plot->paintapi, layout);

      //printf("%d,%d %d,%d '%s'\n", ri.xl, ri.yt, ri.xr, ri.yb, example_time);

      plot->time_stamp_width = ri.xr-ri.xl+1;

      //printf("tsw %s %d\n", example_time, plot->time_stamp_width);

      vgrid = calc_time_grid(plot->xmin, plot->xmax, (int)(plot->plot_xe / (plot->time_stamp_width * (1.3 / 2))), plot->plot_xe + 1);

      min = vgrid.dvmin;

      pprintf("%lf %lf %lf %lf %lf %lf %d\n", vgrid.dmin, vgrid.d, vgrid.dmax, vgrid.dvmin, vgrid.dv, vgrid.dvmax, vgrid.minpermaj);

      time_field_t resulting_min_diff_field = time_diff_field((my_time_t)(vgrid.dv * (1 + 1e-10)));

      if(resulting_min_diff_field == TIME_FIELD_HOUR) resulting_min_diff_field = TIME_FIELD_MIN;

      //printf("pass %d - %d\n", resulting_min_diff_field, min_diff_field);

      if(resulting_min_diff_field <= min_diff_field) {
	 if(resulting_min_diff_field < min_diff_field) {
	    vgrid = calc_time_grid(plot->xmin, plot->xmax, (int)((plot->xmax - plot->xmin) / (time_field_length[min_diff_field] * (1.3 / 2))), plot->plot_xe + 1);
	    resulting_min_diff_field = time_diff_field((my_time_t)(vgrid.dv * (1 + 1e-10)));
	    //printf("updated %d / %d\n", resulting_min_diff_field, min_diff_field);
	    pprintf("%lf %lf %lf %lf %lf %lf %d\n", vgrid.dmin, vgrid.d, vgrid.dmax, vgrid.dvmin, vgrid.dv, vgrid.dvmax, vgrid.minpermaj);
	 }
	 break;
      }
   }

   //printf("--> minmax %d %d\n", min_diff_field, max_diff_field);

   ////

   int font_h = ri.yb - ri.yt + 1;

   plot->font_height = font_h;
   plot->plot_yde = plot->plot_ye - (2 + font_h * 2.2);

   ////

   plot->paintapi->draw_line(plot->paintapi, plot->sep_gc, 0, plot->plot_yde+1, plot->plot_xe, plot->plot_yde+1);

   // set dash offsets
   double ylength = plot->ymax - plot->ymin;
   int doff = -(int)(plot->ymin * plot->plot_yde / ylength);
   plot->paintapi->gc_set_dashes(plot->paintapi, plot->major_gc, doff, major_dashes, SZ(major_dashes));
   plot->paintapi->gc_set_dashes(plot->paintapi, plot->minor_gc, doff, minor_dashes, SZ(minor_dashes));

   int odd = ((int)round(vgrid.dvmin / vgrid.dv)) % 2;
   double xo,xvo;
   for(xo = vgrid.dmin, xvo = vgrid.dvmin; xo < vgrid.dmax + vgrid.d; xo += vgrid.d, xvo += vgrid.dv) {
      odd = !odd;
      int x = (int)xo;

      const char *text = format_time2((my_time_t)round(xvo), min_diff_field, max_diff_field);
      paintapi_textlayout_t *layout = plot->paintapi->textlayout_create(plot->paintapi, plot->font_xaxis, text);

      paintapi_textlayout_extents_t ri;
      plot->paintapi->textlayout_calculate_size(plot->paintapi, layout, &ri);

      int y = plot->plot_yde + 1 + 2 - ri.yt + (int)(plot->font_height * (odd ? 0.1 : 1.2));

      plot->paintapi->draw_textlayout(plot->paintapi, plot->label_gc, x - (ri.xr-ri.xl+1)/2, y, layout);

      plot->paintapi->textlayout_free(plot->paintapi, layout);

      paintapi_gc_t *gc = ((int)round(xvo / vgrid.dv)) % vgrid.minpermaj == 0 ? plot->major_gc : plot->minor_gc;

      plot->paintapi->draw_line(plot->paintapi, gc, x, 0, x, y - 2);
   }
}

void plot_draw_lines(my_plot_draw_t *plot) {
   my_time_t start_time = (my_time_t)plot->xmin;
   my_time_t end_time = (my_time_t)plot->xmax;

   double xlength = plot->xmax - plot->xmin;
   double ylength = plot->ymax - plot->ymin;

   // max points per native paint command
#define MAX_POINTS 1000

   double w_div_xlength = plot->plot_xe / xlength;
   double xlength_div_w_ppx = xlength / (plot->plot_xe * plot->samples_per_pixel);
   double h_div_vdiff = plot->plot_yde / ylength;

   //GdkRectangle clip = { 0, 0, plot->plot_xe, plot->plot_yde }; TODO paintapi

   // allocate points array fulfilling the maximum need
   int maxlines = 0;
   int lsi;
   for(lsi = 0; lsi < plot->line_set_count; ++lsi) {
      maxlines = max(maxlines, plot->line_set[lsi].line_count);
   }
   paintapi_point_t *points = malloc_c(maxlines * MAX_POINTS * sizeof(paintapi_point_t));
   int *usedPoints = malloc_c(maxlines * sizeof(int));
   double *y = malloc_c(maxlines * sizeof(double));
   boolean_t *gotundef = malloc_c(maxlines * sizeof(boolean_t));

   for(lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_dataset_t *dataset = plot->line_set[lsi].dataset;
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;

      int start_index = my_time_arr_find(dataset->time, start_time, 0);
      int end_index = my_time_arr_find(dataset->time, end_time, 1);

      //	if(gtk_events_pending()) {
      //	    $da->queue_draw() unless($queued++);
      //
      //	    $conc--; return;
      //	}

      int li;
      for(li=0; li<line_count; ++li) {
	 if(!line[li].enabled) continue;
	 //#TODO## rectangle ref
	 // plot->paintapi->gc_set_clip_rectangle(plot->paintapi, line[li].line_gc, &clip); TODO paintapi
	 line[li].data = dataset->item[line[li].dataset_idx].data;
	 pprintf("[%d] = %lf\n", li, dataset->item[li].max);
      }

      double x = dataset->time[start_index] - plot->xmin;

      // how many data entries can be skipped per row. Starts from 1, approaches correct value during rendering
      int skip_count = 1;

      bzero(usedPoints, line_count * sizeof(int));
      bzero(gotundef, line_count * sizeof(boolean_t));

      pprintf("ii %d %d\n", start_index, end_index);

      int ti;
      for(ti=start_index ; likely(ti<=end_index); ) {
	 double start_x = x;

	 // ceil is more precise but 3-10% slower - the end result is we draw a little less points than requested
	 //double next_x = ceil(x * w_ppx_div_xlength) * xlength_div_w_ppx;
	 double next_x = x + xlength_div_w_ppx;

	 bzero(y, line_count * sizeof(double));
	 int start_ti = ti;
	 int span_count = 0;
	 while(1) {
	    for(li=0; li<line_count; ++li) {
	       if(!line[li].enabled) continue;
	       double data = line[li].data[ti];
	       if(data > 84) {
		  pprintf("t%d i%d %lf\n", ti, li, data);
	       }
	       if(!__isnan(data)) {
		  switch(line[li].mode) {
		     case LINE_MODE_NORMAL:
			y[li] += data;
			break;
		     case LINE_MODE_DIFF:
			if(span_count == 0)
			   y[li] = data;
			break;
		  }
	       } else {
		  gotundef[li] = 1;
	       }
	    }
	    ti += skip_count;
	    span_count++;
	    if(ti > end_index) break;
	    x = dataset->time[ti] - plot->xmin; // ##TODO## this sometimes gives undef value in subtract
	    if(x > next_x) break;
	 }

	 // adjust skip value
	 skip_count = span_count * skip_count / 4;
	 if(skip_count < 1) skip_count = 1;

	 if((dataset->time[start_ti] - dataset->time[ti]) / (start_ti - ti) > dataset->maxdiff) {
	    memset(gotundef, 1, line_count * sizeof(boolean_t));
	 }

	 int li;
	 for(li=0; li<line_count; ++li) {
	    if(!line[li].enabled) continue;
	    if(gotundef[li]) {
	       gotundef[li] = 0;
	       //pprintf("gun %d %d\n", li, usedPoints[li]);
	       if(usedPoints[li] > 1) {
		  pprintf("gu %d %d\n", li, usedPoints[li]);
		  plot->paintapi->draw_lines(plot->paintapi, line[li].line_gc, &points[MAX_POINTS * li], usedPoints[li]);
	       }
	       usedPoints[li] = 0;
	    } else {
	       double yv;
	       switch(line[li].mode) {
		  case LINE_MODE_NORMAL:
		     yv = span_count > 1 ? y[li] / span_count : y[li];
		     break;
		  case LINE_MODE_DIFF: {
		     int end_ti = ti < end_index ? ti : end_index;
//		     yv = (line[li].data[end_ti] - y[li]) * TIME_MULTIPLIER / (dataset->time[end_ti] - dataset->time[start_ti]);
		     yv = (line[li].data[end_ti] - y[li]) * TIME_MULTIPLIER / (dataset->time[end_ti] - dataset->time[start_ti]) * (1 * TIME_MULTIPLIER);
		     break;
		     default:
			fprintf(stderr, "Bad mode %d in %d.%d\n", line[li].mode, lsi, li);
			return;
		  }
	       }

	       // calculate point coordinates
	       paintapi_point_t *point = &points[MAX_POINTS * li + usedPoints[li]++]; //##TODO##
	       point->x = start_x * w_div_xlength;
	       int ly = (plot->ymax - yv) * h_div_vdiff;

	       // with extreme overdrives, these tend to go wrong (or so I guess)
	       if(ly > 32767) ly = 32767;
	       else if(ly < -32768) ly = -32768;
	       point->y = ly;

	       if(usedPoints[li] == MAX_POINTS) {
		  pprintf("gm %d %d\n", li, usedPoints[li]);
		  plot->paintapi->draw_lines(plot->paintapi, line[li].line_gc, &points[MAX_POINTS * li], usedPoints[li]);
		  points[MAX_POINTS * li] = points[MAX_POINTS * li + usedPoints[li]-1];
		  usedPoints[li] = 1;
	       }
	    }
	 }
      }

      // plot final lines
      for(li=0; li<line_count; ++li) {
	 if(!line[li].enabled) continue;
	 if(usedPoints[li] > 1) {
	    pprintf("f  %d %d\n", li, usedPoints[li]);
	    plot->paintapi->draw_lines(plot->paintapi, line[li].line_gc, &points[MAX_POINTS * li], usedPoints[li]);
	 }
      }

      pprintf("skip_count = %d\n", skip_count);

   }

   free(points);
   free(usedPoints);
   free(y);
   free(gotundef);
}

void plot_draw_legend(my_plot_draw_t *plot) {

   if(plot->show_legend == LEGEND_HIDE) return;

   // precalculate layouts to get max layout width
   int layout_max_x = 0;
   int layout_max_y = 0;
   int layout_min_y = 1000;

   int list_count = 0;
   int lsi;
   for(lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_dataset_t *dataset = plot->line_set[lsi].dataset;
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;

      int li;
      for(li=0; li<line_count; ++li) {
	 if(!line[li].enabled && !plot->show_disabled) continue;

	 const char *linename = dataset->item[line[li].dataset_idx].name;
	 char *text = malloc((dataset->name ? strlen(dataset->name) : 7 + 1 + 10) + 2 + (linename ? strlen(linename) : 4 + 1 + 10) + (line[li].mode == LINE_MODE_DIFF ? 7 : 0) + 1);

	 if(plot->line_set_count > 1) {
	    if(dataset->name) {
	       strcpy(text, dataset->name);
	       strcat(text, ": ");
	    } else {
	       sprintf(text, "dataset %d: ", lsi + 1);
	    }
	 } else {
	    text[0] = '\0';
	 }

	 if(linename) {
	    strcat(text, linename);
	 } else {
	    sprintf(text + strlen(text), "item %d", li + 1);
	 }

	 switch(line[li].mode) {
	    case LINE_MODE_NORMAL:
	       break;

	    case LINE_MODE_DIFF:
	       strcat(text, " (diff)");
	       break;
	 }

	 paintapi_textlayout_t *layout = plot->paintapi->textlayout_create(plot->paintapi, plot->font_legend, text);

	 // overstrike if not enabled
	 if(!line[li].enabled) {
	    plot->paintapi->textlayout_set_strikeout(plot->paintapi, layout);
	 }

	 paintapi_textlayout_extents_t ri;
	 plot->paintapi->textlayout_calculate_size(plot->paintapi, layout, &ri);

	 layout_max_x = max(layout_max_x, ri.xr);
	 layout_max_y = max(layout_max_y, ri.yb);
	 layout_min_y = min(layout_min_y, ri.yt);

	 pprintf("h %d - %d\n", ri.yt, ri.yb);

	 free(text);

	 line[li].legend_layout = layout;

	 list_count++;
      }
   }

   if(list_count == 0) return;

   pprintf("max %d,%d min %d newmax %d\n", layout_max_x, layout_max_y, layout_min_y, layout_max_y - layout_min_y);

   int layout_max_width = layout_max_x + 1;
   int layout_max_height = layout_max_y - layout_min_y + 1;

   pprintf("\n");

   double box_size_multiplier = 0.5;

   int legend_list_box_offset_x = 0;
   int legend_list_box_offset_y = (layout_max_height - 2 - (int)(layout_max_height * box_size_multiplier)) / 2;
   int legend_list_box_size_x = (int)(layout_max_height * box_size_multiplier);
   int legend_list_box_size_y = (int)(layout_max_height * box_size_multiplier);
   int legend_list_box_text_spacing_x = 2 + legend_list_box_size_x / 2;

   // printf("%d %d %d\n", legend_list_box_offset_y, legend_list_box_size_y, layout_max_height);

   int legend_list_unit_size_x = layout_max_width + legend_list_box_offset_x + legend_list_box_size_x + legend_list_box_text_spacing_x;
   int legend_list_unit_size_y = layout_max_height;

   int legend_list_size_x = legend_list_unit_size_x;
   int legend_list_size_y = legend_list_unit_size_y * list_count;

   int legend_margin_left_x = plot->value_axis_width * 4 / 3 + 10;
   int legend_margin_right_x = 10;
   int legend_margin_y = 10;

   int legend_list_padding_x = 4 + legend_list_box_size_x / 2;
   int legend_list_padding_y = 5;

   int legend_corner_x =
      plot->show_legend == LEGEND_SHOW_LEFT ?
      legend_margin_left_x :
      plot->plot_xe - legend_margin_right_x - 2 * legend_list_padding_x - legend_list_size_x;
   int legend_corner_y = legend_margin_y;

   int legend_corner_x2 = legend_corner_x + 2 * legend_list_padding_x + legend_list_size_x;
   int legend_corner_y2 = legend_corner_y + 2 * legend_list_padding_y + legend_list_size_y;

   // draw box
   plot->paintapi->draw_rectangle(plot->paintapi, plot->legend_bg_gc,     1, legend_corner_x+1, legend_corner_y+1, legend_corner_x2-1, legend_corner_y2-1);
   plot->paintapi->draw_rectangle(plot->paintapi, plot->legend_border_gc, 0, legend_corner_x  , legend_corner_y  , legend_corner_x2  , legend_corner_y2  );

   plot->legend_x = legend_corner_x;
   plot->legend_y = legend_corner_y;
   plot->legend_x2 = legend_corner_x2;
   plot->legend_y2 = legend_corner_y2;

   int no = 0;

   // draw legend items
   for(lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;

      int li;
      for(li=0; li<line_count; ++li) {
	 if(!line[li].enabled && !plot->show_disabled) {
	    line[li].legend_x = line[li].legend_y = line[li].legend_x2 = line[li].legend_y2 = -1;
	    continue;
	 }

	 int legend_list_corner_x = legend_corner_x + legend_list_padding_x;
	 int legend_list_corner_y = legend_corner_y + legend_list_padding_y + legend_list_unit_size_y * no;

	 line[li].legend_x = legend_list_corner_x;
	 line[li].legend_y = legend_list_corner_y;
	 line[li].legend_x2 = legend_list_corner_x + legend_list_unit_size_x;
	 line[li].legend_y2 = legend_list_corner_y + legend_list_unit_size_y;

	 int legend_list_text_corner_x = legend_list_corner_x + legend_list_box_offset_x + legend_list_box_size_x + legend_list_box_text_spacing_x;
	 int legend_list_text_corner_y = legend_list_corner_y;

	 plot->paintapi->draw_rectangle(plot->paintapi, line[li].enabled ? line[li].line_gc : plot->legend_border_gc, line[li].enabled,
			      legend_list_corner_x + legend_list_box_offset_x,
			      legend_list_corner_y + legend_list_box_offset_y,
			      legend_list_corner_x + legend_list_box_offset_x + legend_list_box_size_x - 1,
			      legend_list_corner_y + legend_list_box_offset_y + legend_list_box_size_y - 1);

/*
	 plot->paintapi->gc_set_function(plot->paintapi, plot->legend_border_gc, PAINTAPI_FUNCTION_XOR);
	 plot->paintapi->draw_rectangle(plot->paintapi, plot->legend_border_gc, 0,
			      legend_list_corner_x, legend_list_corner_y,
			      legend_list_corner_x + legend_list_unit_size_x - 1, legend_list_corner_y + legend_list_unit_size_y - 1);
	 plot->paintapi->gc_set_function(plot->paintapi, plot->legend_border_gc, PAINTAPI_FUNCTION_SET);
*/

	 plot->paintapi->draw_textlayout(plot->paintapi, plot->legend_text_gc,
			 legend_list_text_corner_x,
			 legend_list_text_corner_y - layout_min_y, line[li].legend_layout);

	 plot->paintapi->textlayout_free(plot->paintapi, line[li].legend_layout);

/*
	 if(!line[li].enabled) {
	    printf("%d %d\n", lx, ly);
	    plot->paintapi->draw_line(plot->paintapi, plot->legend_text_gc,
			  legend_list_text_corner_x, legend_list_text_corner_y + ly/2,
			  legend_list_text_corner_x + lx, legend_list_text_corner_y + ly/2);
	 }
*/

	 no++;
      }
   }

   //    last outer2 if(normal_expose);

   //	    if(mytrans != plot->dsc_x->{num_changes} + plot->dsc_y->{num_changes}) {
   //	    print "- mytrans plot->dsc_x->{num_changes}\n";
   //	    da->queue_draw() unless(queued++);
   //	    queued++;
   //	    plot->conc--; return;

   //####################
}

void plot_draw(my_plot_draw_t *plot) {
   // check settings
   if(unlikely(plot->plot_xe<=0 || plot->plot_ye<=0 || !plot->paintapi || plot->samples_per_pixel<=0)) {
      fprintf(stderr, "You have not set up my_plot_draw_t correctly!\n"
	      "See plot_draw() in plot_draw.xc and comments in plot_draw.xh for help. Asserts and (values):\n"
	      "\tplot->plot_xe>0 (%d)\n"
	      "\tplot->plot_ye>0 (%d)\n"
	      "\tplot->paintapi (%p)\n"
	      "\tplot->samples_per_pixel>0 (%lf)\n",
	      plot->plot_xe,
	      plot->plot_ye,
	      plot->paintapi,
	      plot->samples_per_pixel);
      return;
   }

   struct timeval start;
   gettimeofday(&start, 0);

   // setup gc
   plot_setup_gcs(plot);

   // my $queued = 0;

   //    gtk_main_iteration() while(gtk_events_pending());
   //    if($mytrans != plot->dsc_x->{num_changes}) {
   //	print "- $mytrans plot->dsc_x->{num_changes}\n";
   //#	    $da->queue_draw() unless($queued++);
   //	$conc--; return;
   //    }

   //############# grids

   // time grid
   plot_draw_time_grid(plot);

   // value grid
   plot_draw_value_grid(plot);

   //############ data

   if(system_check_events()) return;

   plot_draw_lines(plot);

   if(system_check_events()) return;

   plot_draw_legend(plot);

   struct timeval end;
   gettimeofday(&end, 0);

   plot->last_render_time = (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec) * 0.000001;
}

void plot_draw_init(my_plot_draw_t *plot) {
   bzero(plot, sizeof(my_plot_draw_t));
   plot->line_set_dflt.line_dflt.hline = -1;
   plot->line_set_dflt.line_dflt.dataset_idx = -1;
   plot->samples_per_pixel = 4;
}

void plot_draw_deinit(my_plot_draw_t *plot) {
   plot_draw_remove_all_data_files(plot);

   plot->paintapi->gc_free(plot->paintapi, plot->minor_gc);
   plot->paintapi->gc_free(plot->paintapi, plot->major_gc);
   plot->paintapi->gc_free(plot->paintapi, plot->sep_gc);
   plot->paintapi->gc_free(plot->paintapi, plot->label_gc);
   plot->paintapi->gc_free(plot->paintapi, plot->cursor_marker_gc);

   plot->paintapi->api_free(plot->paintapi);

   bzero(plot, sizeof(my_plot_draw_t));
}

static void plot_draw_free_line_info(my_plot_draw_t *plot, my_plot_line_info_t *p) {
   plot->paintapi->gc_free(plot->paintapi, p->line_gc);
   plot->paintapi->gc_free(plot->paintapi, p->horiz_marker);
}

static void plot_draw_free_line_info_set(my_plot_draw_t *plot, my_plot_line_info_set_t *p) {
   int li;
   for(li=0; li<p->line_count; ++li) {
      plot_draw_free_line_info(plot, &p->line[li]);
   }
   my_plot_line_info_arr_free(p->line);

   unref_dataset(p->dataset);
}

void plot_draw_remove_all_data_files(my_plot_draw_t *plot) {
   int lsi;
   for(lsi=0; lsi<plot->line_set_count; ++lsi) {
      plot_draw_free_line_info_set(plot, &plot->line_set[lsi]);
   }
   my_plot_line_info_set_arr_free(plot->line_set);
}

void plot_draw_remove_data_file(my_plot_draw_t *plot, int idx) {
   plot_draw_free_line_info_set(plot, &plot->line_set[idx]);
   --plot->line_set_count;
   memmove(&plot->line_set[idx], &plot->line_set[idx+1], (plot->line_set_count - idx) * sizeof(my_plot_line_info_set_t));
}

#define gc_free_and_zero(x) do { plot->paintapi->gc_free(plot->paintapi, x); x = NULL; } while(0)

void plot_draw_reset_gc(my_plot_draw_t *plot) {
   int lsi;
   for(lsi=0; lsi<plot->line_set_count; ++lsi) {
      my_plot_line_info_set_t *pp = &plot->line_set[lsi];
      int li;
      for(li=0; li<pp->line_count; ++li) {
	 my_plot_line_info_t *p = &pp->line[li];
	 if(p->line_gc) gc_free_and_zero(p->line_gc);
      }
   }

   gc_free_and_zero(plot->minor_gc);
   gc_free_and_zero(plot->major_gc);
   gc_free_and_zero(plot->sep_gc);
   gc_free_and_zero(plot->label_gc);
   gc_free_and_zero(plot->cursor_marker_gc);
}
